
/*
 *  calcengine.cpp
 *  Ulambator v0.3
 *
 *  Created by Mathias Nagel on 29.01.10.
 *  Copyright 2010 EPFL. All rights reserved.
 *
 */
#include <iostream>
#include <fstream>
#include "bndelement.h"
#include "matblc1.h"
#ifdef _WIN32
#define _USE_MATH_DEFINES // for C++
#include <cmath>
#endif
#include <math.h>
#include "combase.h"
#include <iomanip>
#ifdef _WIN32
#include <algorithm>
#endif

/*
    Matrix Class add on, this contains all the Matrix writing prcedure
    for historic reasons its outside matrixblock because flexibility
    for higher order routines was intended.
    Although this strategy is discarded the separation is still usefull for readability
 
    Contents:
    - Matrix writing
    - Matrix writing for precondensed problems
    - Visual writing

*/

MatBlC1::MatBlC1(BndBlock *bembl, double aspectRatio, double viscRatio, double capillaryNumber)
: MatrixBlock(bembl, aspectRatio, viscRatio, capillaryNumber)
{
}


// Build blocks for Matrix construction based on respective boundary condition

void MatBlC1::ZeroVel(int inP){
	
	/*	Integration from all singularities over a boundary with zero velocity.
		Wall Elements are always zeroth order	
	 */
	
	double dx, dy, xls, yls, xgs, ygs, r;
	double x1, y1, gvar, G11, G12, G22;
    double *Xp, *Yp, *Xs, *Ys;  // Xp for panels and Xs for singularities
	int localElm = bem->allSize[inP];
	int rmtElms, mp1, fixind, dolog;
	int xMatID = 0;
	int yMatID = 0;
    
	Xp = bem->allXId[inP];
    Yp = bem->allYId[inP];
    
	for(int m=0; m<inP; m++){
		xMatID += 2*bem->allSize[m];
	}
	
	//	Size of the Element, the Position of its center, Normals
	for(int n=0; n<localElm; n++){
		dx = Xp[n+1]-Xp[n];					
        dy = Yp[n+1]-Yp[n];
        xls = 0.5*(Xp[n+1]+Xp[n]);
        yls = 0.5*(Yp[n+1]+Yp[n]);
        r = (dx*dx+dy*dy);
        r = sqrt(r);

		yMatID = 0;
		mp1=1;
        
        A[(2*n+xMatID)*(nOFm+1)] += -r*(log(r/2)-1);
        A[(2*n+1+xMatID)*(nOFm+1)] += -r*(log(r/2)-1);
		
		//	Loop over all panels, switch for singularity position (midpoint corner)
		for(int k=0; k<nbBlocks; k++){
			rmtElms = bem->Elements[k].getFullSize();
            Xs = bem->Elements[k].XId[0];
            Ys = bem->Elements[k].YId[0];
            fixind = yMatID+nOFm*xMatID+2*n*nOFm;
			mp1 = (k<bem->ifblock);
			
			#pragma omp parallel
			{
				#pragma omp for private(dolog,xgs,ygs,x1,y1,G11,G12,G22,gvar)
				//	Loop over all elements of a panel
				for(int m=0; m<rmtElms; m++){
                    
					xgs = xls-0.5*(Xs[m+mp1]+Xs[m]);
					ygs = yls-0.5*(Ys[m+mp1]+Ys[m]);
				
					gvar = xgs*xgs+ygs*ygs;
                    
                    // Decide wheather it is a singular element				
                    dolog = (gvar>1e-12);
                        
                        // Loop over Gauss points
                    for(int l=0; l<6; l++){
                        x1 = 0.5*dx*IGP[l]+xgs;
                        y1 = 0.5*dy*IGP[l]+ygs; 
                            
                        computeGO(dolog, x1, y1, G11, G12, G22);
                        gvar = IGW[l]*r*0.5;
                            
                        A[fixind+2*m] -= G11*gvar;
                        A[fixind+nOFm+2*m] -= G12*gvar;
                        A[fixind+1+2*m] -= G12*gvar;
                        A[fixind+nOFm+1+2*m] -= G22*gvar;
                    }
				}
			}
            # pragma omp barrier
			yMatID += 2*rmtElms;
		}
	}
}


void MatBlC1::FlatVel(int inP, double Vn, double Vt, double Theta, double dEta){
	
	/*	Integration from all singularities over a boundary with n-t velocity.
	 Wall Elements are always zeroth order	
	 */
	
	double dx, dy, xls, yls, xgs, ygs, gvar, r, lnx, lny, u1, u2, xcm, ycm;
	double x1, y1, G11, G12, G22, T111, T122, T112, T211, T222, T212;
	
	double *Xp, *Yp, *Xs, *Ys;  // Xp for panels and Xs for singularities
	int localElm = bem->allSize[inP];
	int mp1 = 1;
	int rmtElms, dolog, fixind;
	int xMatID = 0;
	int yMatID = 0;
	
	Xp = bem->allXId[inP];
    Yp = bem->allYId[inP];
    
	for(int m=0; m<inP; m++){
		xMatID += 2*bem->allSize[m];
	}
    
    bem->getCoM(bem->panel2block[inP], xcm, ycm);
	
	//	Size of the Element, the Position of its center, normals
	for(int n=0; n<localElm; n++){
		dx = Xp[n+1]-Xp[n];					
        dy = Yp[n+1]-Yp[n];
        xls = 0.5*(Xp[n+1]+Xp[n]);
        yls = 0.5*(Yp[n+1]+Yp[n]);
        r = (dx*dx+dy*dy);
        r = sqrt(r);
        lnx = dy/r;
        lny = -dx/r;
        u1 = Vn - (yls-ycm)*Theta;
        u2 = Vt + (xls-xcm)*Theta;
		yMatID = 0;
		mp1=1;
        
        A[(2*n+xMatID)*(nOFm+1)] -= r*(log(r/2)-1);
        A[(2*n+1+xMatID)*(nOFm+1)]  -= r*(log(r/2)-1);
        
        rhs[xMatID+2*n] += 2*dEta*M_PI*u1;
        rhs[xMatID+2*n+1] += 2*dEta*M_PI*u2;
        
		//	Loop over all panels, switch for singularity position (midpoint corner)
		for(int k=0; k<nbBlocks; k++){
			rmtElms = bem->Elements[k].getFullSize();
            Xs = bem->Elements[k].XId[0];
            Ys = bem->Elements[k].YId[0];
			mp1 = (k<bem->ifblock);
            fixind = yMatID+nOFm*xMatID+2*n*nOFm;
			#pragma omp parallel
			{
				#pragma omp for private(xgs,ygs,x1,y1,G11,G12,G22,T111,T112,T122,T211,T212,T222,gvar,dolog)
				//	Loop over all elements of a panel
				for(int m=0; m<rmtElms; m++){
					xgs = xls-0.5*(Xs[m+mp1]+Xs[m]);
                    ygs = yls-0.5*(Ys[m+mp1]+Ys[m]);
                    
                    gvar = xgs*xgs+ygs*ygs;
                    
                    // Decide wheather it is a singular element				
                    dolog = (gvar>1e-12);
				
					// Loop over Gauss points
					for(int l=0; l<6; l++){
						x1 = 0.5*dx*IGP[l]+xgs;
                        y1 = 0.5*dy*IGP[l]+ygs; 
						
                        computeGT(dolog, x1, y1, G11, G12, G22, T111, T112, T122, T211, T212, T222);
                        gvar = IGW[l]*r*0.5;
						
                        A[fixind+2*m] -= G11*gvar;
                        A[fixind+nOFm+2*m] -= G12*gvar;
                        A[fixind+1+2*m] -= G12*gvar;
                        A[fixind+nOFm+1+2*m] -= G22*gvar;
						
                        rhs[yMatID+2*m]   -= dEta*((T111*lnx+T112*lny)*u1+(T112*lnx+T122*lny)*u2)*gvar;
                        rhs[yMatID+2*m+1] -= dEta*((T211*lnx+T212*lny)*u1+(T212*lnx+T222*lny)*u2)*gvar;
					}
				}
			}
            # pragma omp barrier
			yMatID += 2*rmtElms;
		}
	}
}

// void MatBlC1::FlatVelrhs(int inP, double Vn, double Vt, double Theta, double dEta, int insert){

void MatBlC1::NormVel(int inP, double Vn, double Vt, double dEta){
	
	/*	Integration from all singularities over a boundary with n-t velocity.
	 Wall Elements are always zeroth order	
	 */
	
	double dx, dy, xls, yls, xgs, ygs, gvar, r, lnx, lny, u1, u2;
	double x1, y1, G11, G12, G22, T111, T122, T112, T211, T222, T212;
	
    double *Xp, *Yp, *Xs, *Ys;  // Xp for panels and Xs for singularities
	int localElm = bem->allSize[inP];
	int mp1 = 1;
	int rmtElms, dolog, fixind;
	int xMatID = 0;
	int yMatID = 0;
	
	Xp = bem->allXId[inP];
    Yp = bem->allYId[inP];
    
	for(int m=0; m<inP; m++){
		xMatID += 2*bem->allSize[m];
	}
    
	//	Size of the Element, the Position of its center, normals
	for(int n=0; n<localElm; n++){
		dx = Xp[n+1]-Xp[n];					
        dy = Yp[n+1]-Yp[n];
        xls = 0.5*(Xp[n+1]+Xp[n]);
        yls = 0.5*(Yp[n+1]+Yp[n]);
        r = (dx*dx+dy*dy);
        r = sqrt(r);
        lnx = dy/r;
        lny = -dx/r;
        u1 = Vn*lnx-Vt*lny;
        u2 = Vn*lny+Vt*lnx;
		yMatID = 0;
		mp1=1;

        A[(2*n+xMatID)*(nOFm+1)] += -r*(log(r/2)-1);
        A[(2*n+1+xMatID)*(nOFm+1)]  += -r*(log(r/2)-1);
        
        rhs[xMatID+2*n] += 2*dEta*M_PI*u1;
        rhs[xMatID+2*n+1] += 2*dEta*M_PI*u2;
     
		//	Loop over all panels, switch for singularity position (midpoint corner)
		for(int k=0; k<nbBlocks; k++){
			rmtElms = bem->Elements[k].getFullSize();
            Xs = bem->Elements[k].XId[0];
            Ys = bem->Elements[k].YId[0];
            fixind = yMatID+nOFm*xMatID+2*n*nOFm;
            mp1 = (k<bem->ifblock);
            
#pragma omp parallel
			{
#pragma omp for private(xgs,ygs,x1,y1,G11,G12,G22,T111,T112,T122,T211,T212,T222,gvar,dolog)
				//	Loop over all elements of a panel
				for(int m=0; m<rmtElms; m++){
					xgs = xls-0.5*(Xs[m+mp1]+Xs[m]);
                    ygs = yls-0.5*(Ys[m+mp1]+Ys[m]);
                    gvar = xgs*xgs+ygs*ygs;
                    
                    // Decide wheather it is a singular element				
                    dolog = (gvar>1e-12);
                    
					// Loop over Gauss points
					for(int l=0; l<6; l++){
						x1 = 0.5*dx*IGP[l]+xgs;
                        y1 = 0.5*dy*IGP[l]+ygs; 
						
                        computeGT(dolog, x1, y1, G11, G12, G22, T111, T112, T122, T211, T212, T222);
                        gvar = IGW[l]*r*0.5;
						
                        A[fixind+2*m] -= G11*gvar;
                        A[fixind+nOFm+2*m] -= G12*gvar;
                        A[fixind+1+2*m] -= G12*gvar;
                        A[fixind+nOFm+1+2*m] -= G22*gvar;
						
                        rhs[yMatID+2*m] -= dEta*((T111*lnx+T112*lny)*u1+(T112*lnx+T122*lny)*u2)*gvar;
                        rhs[yMatID+2*m+1] -= dEta*((T211*lnx+T212*lny)*u1+(T212*lnx+T222*lny)*u2)*gvar;
					}
				}
			}
            # pragma omp barrier
			yMatID += 2*rmtElms;
		}
	}
}

void MatBlC1::POX(int inP, double plevel, double dEta){
	/*	Integration from all singularities over a boundary pressure.
     This boundary works only if the wall is parallel to the y-axis
	 Wall Elements are always zeroth order	
	 */
	
	double dx, dy, xls, yls, xgs, ygs, r, lnx, lny, f1;
	double x1, y1, G11, G12, G22, gvar;
	double T111, T122, T112, T211, T222, T212;
    double *Xp, *Yp, *Xs, *Ys;  // Xp for panels and Xs for singularities
	int localElm = bem->allSize[inP];
	int mp1, dolog, fixind;
	
	int rmtElms;
	int xMatID = 0;
	int yMatID = 0;
	
	Xp = bem->allXId[inP];
    Yp = bem->allYId[inP];
    
	for(int m=0; m<inP; m++){
		xMatID += 2*bem->allSize[m];
	}
	
	//	Size of the Element, the Position of its center, Normals
	for(int n=0; n<localElm; n++){
		dx = Xp[n+1]-Xp[n];					
        dy = Yp[n+1]-Yp[n];
        xls = 0.5*(Xp[n+1]+Xp[n]);
        yls = 0.5*(Yp[n+1]+Yp[n]);
        r = dx*dx+dy*dy;
        r = sqrt(r);
        lnx = dy/r;
        lny = -dx/r;
        f1 = -plevel;
		yMatID = 0;
		mp1=1;

        A[(xMatID+2*n)*(nOFm+1)] += -2*M_PI*dEta;
        A[(xMatID+2*n+1)*(nOFm+1)] += -r*(log(r/2.0)-1);
        rhs[xMatID+2*n] += r*(log(r/2.0)-1)*f1;
		
		//	Loop over all panels, switch for singularity position (midpoint corner)
		for(int k=0; k<nbBlocks; k++){
			rmtElms = bem->Elements[k].getFullSize();
            Xs = bem->Elements[k].XId[0];
            Ys = bem->Elements[k].YId[0];
			mp1 = (k<bem->ifblock);
            fixind = yMatID+nOFm*xMatID+2*n*nOFm;
#pragma omp parallel
			{
#pragma omp for private(xgs,ygs,x1,y1,G11,G12,G22,T111,T112,T122,T211,T212,T222,gvar,dolog)
				//	Loop over all elements of a panel
				for(int m=0; m<rmtElms; m++){
					xgs = xls-0.5*(Xs[m+mp1]+Xs[m]);
                    ygs = yls-0.5*(Ys[m+mp1]+Ys[m]);

                    gvar = xgs*xgs+ygs*ygs;
                    
                    // Decide wheather it is a singular element				
                    dolog = (gvar>1e-14);
					
					// Loop over Gauss points
					for(int l=0; l<6; l++){
						x1 = 0.5*dx*IGP[l]+xgs;
						y1 = 0.5*dy*IGP[l]+ygs; 
						
						computeGT(dolog, x1, y1, G11, G12, G22, T111, T112, T122, T211, T212, T222);
						gvar = IGW[l]*r*0.5;
						
						A[fixind + 2*m] += dEta*(T111*lnx+T112*lny)*gvar;
						A[fixind+nOFm+2*m] -= G12*gvar;
						A[fixind+1+2*m] += dEta*(T211*lnx+T212*lny)*gvar;
						A[fixind+nOFm+1+2*m] -= G22*gvar;
						
						rhs[yMatID+2*m]   += (f1*G11)*gvar;
						rhs[yMatID+2*m+1] += (f1*G12)*gvar;
					}
				}
			}
            # pragma omp barrier
			yMatID += 2*rmtElms;
		}
	}		
}

void MatBlC1::Pressure(int inP, double plevel, double dEta){
	
	/*	Integration from all singularities over a boundary pressure.
	 Wall Elements are always zeroth order	
	*/
	
	double dx, dy, xls, yls, xgs, ygs, r, lnx, lny, f1, f2;
	double x1, y1, G11, G12, G22, gvar;
	double T111, T122, T112, T211, T222, T212;
    double *Xp, *Yp, *Xs, *Ys;  // Xp for panels and Xs for singularities
	int localElm = bem->allSize[inP];
	int mp1, dolog;
	
	int rmtElms, fixind;
	int xMatID = 0;
	int yMatID = 0;
	
	Xp = bem->allXId[inP];
    Yp = bem->allYId[inP];
	
	for(int m=0; m<inP; m++){
		xMatID += 2*bem->allSize[m];
	}
	
	//	Size of the Element, the Position of its center, Normals
	for(int n=0; n<localElm; n++){
		dx = Xp[n+1]-Xp[n];					
        dy = Yp[n+1]-Yp[n];
        xls = 0.5*(Xp[n+1]+Xp[n]);
        yls = 0.5*(Yp[n+1]+Yp[n]);
        r = dx*dx+dy*dy;
        r = sqrt(r);
        lnx = dy/r;
        lny = -dx/r;
        f1 = plevel*lnx;
        f2 = plevel*lny;
		yMatID = 0;
		mp1=1;
        
        A[(2*n+xMatID)*(nOFm+1)] -= 2*M_PI*dEta;
        A[(2*n+1+xMatID)*(nOFm+1)] -= 2*M_PI*dEta;
        rhs[xMatID+2*n] += r*(log(r/2)-1)*f1;
        rhs[xMatID+2*n+1] += r*(log(r/2)-1)*f2;
		
		//	Loop over all panels, switch for singularity position (midpoint corner)
		for(int k=0; k<nbBlocks; k++){
			rmtElms = bem->Elements[k].getFullSize();
            Xs = bem->Elements[k].XId[0];
            Ys = bem->Elements[k].YId[0];
			mp1 = (k<bem->ifblock);
			fixind = yMatID+nOFm*xMatID+2*n*nOFm;
			#pragma omp parallel
			{
				#pragma omp for private(xgs,ygs,x1,y1,G11,G12,G22,T111,T112,T122,T211,T212,T222,gvar,dolog)
				//	Loop over all elements of a panel
				for(int m=0; m<rmtElms; m++){
					xgs = xls-0.5*(Xs[m+mp1]+Xs[m]);
                    ygs = yls-0.5*(Ys[m+mp1]+Ys[m]);
                    
                    gvar = xgs*xgs+ygs*ygs;
                    
                    // Decide wheather it is a singular element				
                    dolog = (gvar>1e-12);	
				
					// Loop over Gauss points
					for(int l=0; l<6; l++){
						x1 = 0.5*dx*IGP[l]+xgs;
						y1 = 0.5*dy*IGP[l]+ygs; 
                        
						computeGT(dolog, x1, y1, G11, G12, G22, T111, T112, T122, T211, T212, T222);
						gvar = IGW[l]*r*0.5;
							
						A[fixind+2*m] += dEta*(T111*lnx+T112*lny)*gvar;
						A[fixind+nOFm+2*m] += dEta*(T112*lnx+T122*lny)*gvar;
						A[fixind+1+2*m] += dEta*(T211*lnx+T212*lny)*gvar;
						A[fixind+nOFm+1+2*m] += dEta*(T212*lnx+T222*lny)*gvar;
							
						rhs[yMatID+2*m]   += (f1*G11+f2*G12)*gvar;
						rhs[yMatID+2*m+1] += (f1*G12+f2*G22)*gvar;
                        
					}
				}
			}
            # pragma omp barrier
			yMatID += 2*rmtElms;
		}
	}		
	
}

void MatBlC1::Poiseuille(int inP, double qdebit, double dEta){
    //  A Poiseuille flow boundary condition, qdebit injects a qdebit mean velocity
	double dx, dy, xls, yls, xgs, ygs, r, lnx, lny, u1, u2;
	double x1, y1, gvar, G11, G12, G22;
	double T111, T122, T112, T211, T222, T212;
	double dxpanel, dypanel, xppanel, yppanel, vel, Vn, dpanel, coshb, pmr;
	double *Xp, *Yp, *Xs, *Ys;  // Xp for panels and Xs for singularities
	int localElm = bem->allSize[inP];

	int mp1 = 1;
	int rmtElms, dolog, fixind;
	int xMatID = 0;
	int yMatID = 0;
	
	Xp = bem->allXId[inP];
    Yp = bem->allYId[inP];
    
	for(int m=0; m<inP; m++){
		xMatID += 2*bem->allSize[m];
	}
	
	// Determine Panel Parameters
	dypanel = Yp[localElm]-Yp[0];
	dxpanel = Xp[localElm]-Xp[0];
	xppanel = 0.5*(Xp[localElm]+Xp[0]);
	yppanel = 0.5*(Yp[localElm]+Yp[0]);
	dpanel = sqrt(dxpanel*dxpanel+dypanel*dypanel);
//	nxpanel = dypanel/dpanel;
//	nypanel = -dxpanel/dpanel;
	coshb = cosh(kF*dpanel*0.5);
	vel = qdebit*dpanel*kF/(dpanel*kF*coshb-2*sinh(kF*dpanel/2));
	
	//	Size of the Element, the Position of its center, normals
	for(int n=0; n<localElm; n++){
		dx = Xp[n+1]-Xp[n];					
        dy = Yp[n+1]-Yp[n];
        xls = 0.5*(Xp[n+1]+Xp[n]);
        yls = 0.5*(Yp[n+1]+Yp[n]);
        r = (dx*dx+dy*dy);
        r = sqrt(r);
        lnx = dy/r;
        lny = -dx/r;
		pmr = sqrt((xls-xppanel)*(xls-xppanel)+(yls-yppanel)*(yls-yppanel));
		if(kF!=0){		
			Vn = vel*(coshb-cosh(kF*pmr));
		} else {
			Vn = qdebit/dpanel*1.5*(1.0-4*(pmr/dpanel)*(pmr/dpanel));
		}
		u1 = -Vn*lnx;       // Minus because inflow is in opposite direction of normal
        u2 = -Vn*lny;
		yMatID = 0;
		mp1=1;
        
        A[(2*n+xMatID)*(nOFm+1)] += -r*(log(r/2)-1);
        A[(2*n+1+xMatID)*(nOFm+1)] += -r*(log(r/2)-1);
        
        rhs[xMatID+2*n] += 2*dEta*M_PI*u1;
        rhs[xMatID+2*n+1] += 2*dEta*M_PI*u2;

		//	Loop over all panels, switch for singularity position (midpoint corner)
		for(int k=0; k<nbBlocks; k++){
			rmtElms = bem->Elements[k].getFullSize();
            Xs = bem->Elements[k].XId[0];
            Ys = bem->Elements[k].YId[0];
            fixind = yMatID+nOFm*xMatID+2*n*nOFm;
			mp1 = (k<bem->ifblock);
            
			#pragma omp parallel
			{
				#pragma omp for private(xgs,ygs,x1,y1,G11,G12,G22,T111,T112,T122,T211,T212,T222,gvar,dolog)
			
				//	Loop over all elements of a panel
				for(int m=0; m<rmtElms; m++){
					xgs = xls-0.5*(Xs[m+mp1]+Xs[m]);
                    ygs = yls-0.5*(Ys[m+mp1]+Ys[m]);
                    
                    gvar = xgs*xgs+ygs*ygs;
                    
                    // Decide wheather it is a singular element				
                    dolog = (gvar>1e-12);
				
					// Loop over Gauss points
					for(int l=0; l<6; l++){
						x1 = 0.5*dx*IGP[l]+xgs;
						y1 = 0.5*dy*IGP[l]+ygs; 
					
						computeGT(dolog, x1, y1, G11, G12, G22, T111, T112, T122, T211, T212, T222);
						gvar = IGW[l]*r*0.5;
						
						A[fixind+2*m] -= G11*gvar;
                        A[fixind+nOFm+2*m] -= G12*gvar;
                        A[fixind+1+2*m] -= G12*gvar;
                        A[fixind+nOFm+1+2*m] -= G22*gvar;
						
						rhs[yMatID+2*m] -= dEta*((T111*lnx+T112*lny)*u1+(T112*lnx+T122*lny)*u2)*gvar;
						rhs[yMatID+2*m+1] -= dEta*((T211*lnx+T212*lny)*u1+(T212*lnx+T222*lny)*u2)*gvar;
                        
					}
				}
			}
            # pragma omp barrier
			yMatID += 2*rmtElms;
		}
	}
}

void MatBlC1::InfiniBox(double xvel, double yvel){
    // velocity at infinity
    int size=0;
    int counter = 0;
    
    for (int k=0; k<nbBlocks; k++) {
        size = bem->Elements[k].getFullSize();
        
        for (int j=0; j<size; j++) {
            rhs[2*counter] += -4.0*M_PI*xvel;
            rhs[2*counter+1] += -4.0*M_PI*yvel;
            counter++;
        }
    }
}
void MatBlC1::InfiniLinearFlow(double lin, double colin, double dEta){
    // linear flow at infinity
    int size=0;
    double *Xp, *Yp;
    int counter = 0;
    
    for (int k=0; k<nbBlocks; k++) {
        size = bem->Elements[k].getFullSize();
        Xp = bem->Elements[k].XId[0];
        Yp = bem->Elements[k].YId[0];
        for (int n=0; n<size; n++) {
            rhs[2*counter] += -4*M_PI*(lin*Xp[n] + colin*Yp[n]);
            rhs[2*counter+1] += -4*M_PI*(-lin*Yp[n] + colin*Xp[n]);
            counter++;
        }
    }
}

void MatBlC1::FiniLinearFlow(int inP, double lin, double colin, double dEta){
    // linear flow on a domain boundary, used to test infinilinearflow, basically useless
	double dx, dy, xls, yls, xgs, ygs, gvar, r, lnx, lny, u1, u2;
	double x1, y1, G11, G12, G22, T111, T122, T112, T211, T222, T212;
	
	double *Xp, *Yp, *Xs, *Ys;  // Xp for panels and Xs for singularities
	int localElm = bem->allSize[inP];
	int mp1 = 1;
	int rmtElms, dolog, fixind;
	int xMatID = 0;
	int yMatID = 0;
	
	Xp = bem->allXId[inP];
    Yp = bem->allYId[inP];
    
	for(int m=0; m<inP; m++){
		xMatID += 2*bem->allSize[m];
	}
	
	//	Size of the Element, the Position of its center, normals
	for(int n=0; n<localElm; n++){
		dx = Xp[n+1]-Xp[n];					
        dy = Yp[n+1]-Yp[n];
        xls = 0.5*(Xp[n+1]+Xp[n]);
        yls = 0.5*(Yp[n+1]+Yp[n]);
        r = (dx*dx+dy*dy);
        r = sqrt(r);
        lnx = dy/r;
        lny = -dx/r;
        u1 = xls;
        u2 = -yls;
		yMatID = 0;
		mp1=1;
        
        A[(2*n+xMatID)*(nOFm+1)] += -r*(log(r/2)-1);
        A[(2*n+1+xMatID)*(nOFm+1)]  += -r*(log(r/2)-1);
        
        rhs[xMatID+2*n] += 2*dEta*M_PI*u1;
        rhs[xMatID+2*n+1] += 2*dEta*M_PI*u2;
        
		//	Loop over all panels, switch for singularity position (midpoint corner)
		for(int k=0; k<nbBlocks; k++){
			rmtElms = bem->Elements[k].getFullSize();
            Xs = bem->Elements[k].XId[0];
            Ys = bem->Elements[k].YId[0];
			mp1 = (k<bem->ifblock);
            
#pragma omp parallel
			{
#pragma omp for private(xgs,ygs,x1,y1,G11,G12,G22,T111,T112,T122,T211,T212,T222,gvar,dolog,fixind)
				//	Loop over all elements of a panel
				for(int m=0; m<rmtElms; m++){
					xgs = xls-0.5*(Xs[m+mp1]+Xs[m]);
                    ygs = yls-0.5*(Ys[m+mp1]+Ys[m]);
                    fixind = yMatID+nOFm*xMatID+2*n*nOFm+2*m;
                    gvar = xgs*xgs+ygs*ygs;
                    
                    // Decide wheather it is a singular element				
                    dolog = (gvar>1e-12);
                    
					// Loop over Gauss points
					for(int l=0; l<6; l++){
						x1 = 0.5*dx*IGP[l]+xgs;
                        y1 = 0.5*dy*IGP[l]+ygs; 
						
                        computeGT(dolog, x1, y1, G11, G12, G22, T111, T112, T122, T211, T212, T222);
                        gvar = IGW[l]*r*0.5;
						
                        A[fixind] -= G11*gvar;
                        A[fixind+nOFm] -= G12*gvar;
                        A[fixind+1] -= G12*gvar;
                        A[fixind+nOFm+1] -= G22*gvar;
						
                        rhs[yMatID+2*m]   -= dEta*((T111*lnx+T112*lny)*u1+(T112*lnx+T122*lny)*u2)*gvar;
                        rhs[yMatID+2*m+1] -= dEta*((T211*lnx+T212*lny)*u1+(T212*lnx+T222*lny)*u2)*gvar;
					}
				}
			}
# pragma omp barrier
			yMatID += 2*rmtElms;
		}
	}
}

void MatBlC1::InfiniSource(double qdebit, double xpos, double ypos, double dEta){
    // a source injection
    double *Xp, *Yp;
    int size=0;
    double cr;
    int counter = 0;
    
    for (int k=0; k<nbBlocks; k++) {
        size = bem->Elements[k].getFullSize();
        Xp = bem->Elements[k].XId[0];
        Yp = bem->Elements[k].YId[0];
        
        for (int n=0; n<size; n++) {
            cr = dEta*4*M_PI*qdebit/((Xp[n]-xpos)*(Xp[n]-xpos)+(Yp[n]-ypos)*(Yp[n]-ypos));
            rhs[2*counter] += -cr*Xp[n];   // For radial injection
            rhs[2*counter+1] += -cr*Yp[n];
            counter++;
        }
    }
}

void MatBlC1::SineFlow(int inP, double qAvel, double wavenumber, double dEta){
    //  A Poiseuille flow boundary condition, qdebit injects a qdebit mean velocity
	double dx, dy, xls, yls, xgs, ygs, r, lnx, lny, u1, u2;
	double x1, y1, gvar, G11, G12, G22;
	double T111, T122, T112, T211, T222, T212;
	double dxpanel, dypanel, xppanel, yppanel, Vn, dpanel;
	double *Xp, *Yp, *Xs, *Ys;  // Xp for panels and Xs for singularities
	int localElm = bem->allSize[inP];
    double nxpanel, nypanel;
    double scoord; // global coordinate projected on the straight center line of the panel
    
	int mp1 = 1;
	int rmtElms, dolog, fixind;
	int xMatID = 0;
	int yMatID = 0;
	
	Xp = bem->allXId[inP];
    Yp = bem->allYId[inP];
    
	for(int m=0; m<inP; m++){
		xMatID += 2*bem->allSize[m];
	}
	
	// Determine Panel Parameters
	dypanel = Yp[localElm]-Yp[0];
	dxpanel = Xp[localElm]-Xp[0];
	xppanel = 0.5*(Xp[localElm]+Xp[0]);
	yppanel = 0.5*(Yp[localElm]+Yp[0]);
	dpanel = sqrt(dxpanel*dxpanel+dypanel*dypanel);
    nxpanel = dypanel/dpanel;
    nypanel = -dxpanel/dpanel;
	
    
	
	//	Size of the Element, the Position of its center, normals
	for(int n=0; n<localElm; n++){
		dx = Xp[n+1]-Xp[n];
        dy = Yp[n+1]-Yp[n];
        xls = 0.5*(Xp[n+1]+Xp[n]);
        yls = 0.5*(Yp[n+1]+Yp[n]);
        r = (dx*dx+dy*dy);
        r = sqrt(r);
        lnx = dy/r;
        lny = -dx/r;
		scoord = -(xls-Xp[0])*nypanel + (yls-Yp[0])*nxpanel;
        Vn = qAvel*cos(2*M_PI*wavenumber*scoord/dpanel);
		
		u1 = Vn*nxpanel;       // Minus because inflow is in opposite direction of normal
        u2 = Vn*nypanel;
		yMatID = 0;
		mp1=1;
        
        A[(2*n+xMatID)*(nOFm+1)] += -r*(log(r/2)-1);
        A[(2*n+1+xMatID)*(nOFm+1)] += -r*(log(r/2)-1);
        
        rhs[xMatID+2*n] += 2*dEta*M_PI*u1;
        rhs[xMatID+2*n+1] += 2*dEta*M_PI*u2;
        
		//	Loop over all panels, switch for singularity position (midpoint corner)
		for(int k=0; k<nbBlocks; k++){
			rmtElms = bem->Elements[k].getFullSize();
            Xs = bem->Elements[k].XId[0];
            Ys = bem->Elements[k].YId[0];
            fixind = yMatID+nOFm*xMatID+2*n*nOFm;
			mp1 = (k<bem->ifblock);
            
#pragma omp parallel
			{
#pragma omp for private(xgs,ygs,x1,y1,G11,G12,G22,T111,T112,T122,T211,T212,T222,gvar,dolog)
                
				//	Loop over all elements of a panel
				for(int m=0; m<rmtElms; m++){
					xgs = xls-0.5*(Xs[m+mp1]+Xs[m]);
                    ygs = yls-0.5*(Ys[m+mp1]+Ys[m]);
                    
                    gvar = xgs*xgs+ygs*ygs;
                    
                    // Decide wheather it is a singular element
                    dolog = (gvar>1e-12);
                    
					// Loop over Gauss points
					for(int l=0; l<6; l++){
						x1 = 0.5*dx*IGP[l]+xgs;
						y1 = 0.5*dy*IGP[l]+ygs;
                        
						computeGT(dolog, x1, y1, G11, G12, G22, T111, T112, T122, T211, T212, T222);
						gvar = IGW[l]*r*0.5;
						
						A[fixind+2*m] -= G11*gvar;
                        A[fixind+nOFm+2*m] -= G12*gvar;
                        A[fixind+1+2*m] -= G12*gvar;
                        A[fixind+nOFm+1+2*m] -= G22*gvar;
						
						rhs[yMatID+2*m] -= dEta*((T111*lnx+T112*lny)*u1+(T112*lnx+T122*lny)*u2)*gvar;
						rhs[yMatID+2*m+1] -= dEta*((T211*lnx+T212*lny)*u1+(T212*lnx+T222*lny)*u2)*gvar;
                        
					}
				}
			}
# pragma omp barrier
			yMatID += 2*rmtElms;
		}
	}
}



void MatBlC1::CircTrac(int inP, double delT){
    // The standard droplet boundary condition
	double dx, dy, xls, yls, xgs, ygs, r, lnx, lny;
	double x1, y1, G11, G12, G22, fl1, fl2, fr1, fr2;
	double T111, T122, T112, T211, T222, T212, f1,f2;
	double phil, phir, tx1, ty1, tx3, ty3, s1, s3, d1, d2, unorm, unorp;
	double dLambda = Lambda -1.0;
    double *Xp, *Yp, *Xs, *Ys;  // Xp for panels and Xs for singularities
	int localElm = bem->allSize[inP];
	int rmtElms, nm1, np1, np2, dolog;
//    double fifac = 2*kF*pow(1-Eta,0.66666)*pow(nGamma,0.33333)/(3*sqrt(12));
	int xMatID = 0;
	int yMatID = 0;
    double LMgam, LPgam, RMgam,RPgam,kappa;
    double maxForce = 0;
    double minForce = 1e9;
    
    double alpha = 1.0;
	if(kF < 1e-3){
		alpha=0;
	}
    
    Xp = bem->allXId[inP];
    Yp = bem->allYId[inP];
    
    int fixind;
	
	for(int m=0; m<inP; m++){
		xMatID += 2*bem->allSize[m];
	}
	
	//	Size of the Element, the Position of its center, normals
	for(int n=0; n<localElm; n++){
        
        // indices wrapping the interface
		nm1 = n-1;
		np1 = n+1;
		np2 = n+2;
		if (n==localElm-1) {
			np1 = 0;
			np2 = 1;
		} else if(n==0){
			nm1 = localElm-1;
		}else if(n==localElm-2){
			np2=0;
		} 
		
        // geometrical properties of one element for curvature and integration
		tx1 = Xp[n]-Xp[nm1];
		ty1 = Yp[n]-Yp[nm1];
		s1 = sqrt(tx1*tx1+ty1*ty1);
		tx3 = Xp[np2]-Xp[np1];
		ty3 = Yp[np2]-Yp[np1];
		s3 = sqrt(tx3*tx3+ty3*ty3);
		dx = Xp[np1]-Xp[n];
		dy = Yp[np1]-Yp[n];
		r = sqrt(dx*dx+dy*dy);
		d1 = s1+r;
		d2 = s3+r; 
		
        xls = 0.5*(Xp[n+1]+Xp[n]);
        yls = 0.5*(Yp[n+1]+Yp[n]);
        lnx = dy/r;
        lny = -dx/r;
		
        LMgam = LocalGam(inP, nm1, n)-LocalGam(inP, n, n)*(1.0-M_PI_4)*alpha;
        LPgam = LocalGam(inP, n, np1)-LocalGam(inP, n, n)*(1.0-M_PI_4)*alpha;
        RMgam = LocalGam(inP, n, np1)-LocalGam(inP, np1, np1)*(1.0-M_PI_4)*alpha;
        RPgam = LocalGam(inP, np1, np2)-LocalGam(inP, np1, np1)*(1.0-M_PI_4)*alpha;
        
		fl1 = 2.0*(LPgam*dx/r-LMgam*tx1/s1)/d1;
		fl2 = 2.0*(LPgam*dy/r-LMgam*ty1/s1)/d1;
		fr1 = 2.0*(RPgam*tx3/s3-RMgam*dx/r)/d2;
		fr2 = 2.0*(RPgam*ty3/s3-RMgam*dy/r)/d2;
   
        // model boundary conditions
		unorm = getUvel(inP, nm1, n, np1);
        kappa = LocalGam(inP, n, n); //  2.0/d1*sqrt( pow(dx/r-tx1/s1,2) + pow(dy/r-ty1/s1,2)  );
        (this->*applybc)(Xp[n], Yp[n], lnx, lny, unorm, kappa, f1, f2);
		fl1 +=  f1;
        fl2 +=  f2;
		
		unorp = getUvel(inP, n, np1, np2);
        kappa = LocalGam(inP, np1, np1); //2.0/d2*sqrt( pow(tx3/s3-dx/r,2) + pow(ty3/s3-dy/r,2)  );
        (this->*applybc)(Xp[n+1], Yp[n+1], lnx, lny, unorp, kappa, f1, f2);
		fr1 +=  f1;
        fr2 +=  f2;
 
        // break-up and wall and droplet repulsion
        fl1 += SpecialForces(inP,n)*lnx;
        fl2 += SpecialForces(inP,n)*lny;
        fr1 += SpecialForces(inP,np1)*lnx;
        fr2 += SpecialForces(inP,np1)*lny;
		
        // estimate the magnitude of acting terms to optionally limit the timestep
        maxForce = std::max(sqrt(fl1*fl1+fl2*fl2),maxForce);
        minForce = std::min(sqrt(fl1*fl1+fl2*fl2),minForce);
        
        // index for matrix writing
		yMatID = 0;
        
		A[(2*n+xMatID)*(nOFm+1)] -= 2.0*M_PI*(Lambda + 1.0);
		A[(2*n+1+xMatID)*(nOFm+1)] -= 2.0*M_PI*(Lambda + 1.0);

        // analytic integration of singular terms, two singularities on one element
        rhs[xMatID+2*n]   += (fl1+fr1)*0.25*r*(2*log(r)-1)-0.5*r*fl1;
        rhs[xMatID+2*np1]   += (fl1+fr1)*0.25*r*(2*log(r)-1)-0.5*r*fr1;
   
        rhs[xMatID+2*n+1] += (fl2+fr2)*0.25*r*(2*log(r)-1)-0.5*r*fl2;
        rhs[xMatID+2*np1+1] += (fl2+fr2)*0.25*r*(2*log(r)-1)-0.5*r*fr2;
        
        // analytic integration for surface tension stab.
        phil = -delT*r*0.25*(2*log(r)-3);
        phir = -delT*r*0.25*(2*log(r)-1);
   
        A[xMatID+2*n+nOFm*(2*nm1+xMatID)] += LMgam*phil/(s1*d1);
        A[xMatID+2*np1+nOFm*(2*nm1+xMatID)] += LMgam*phir/(s1*d1);
        
        A[(2*n+xMatID)*(nOFm+1)]          -= phil*(LMgam/(s1*d1)+LPgam/(r*d1))-phir*RMgam/(r*d2);
        A[2*np1+xMatID+nOFm*(xMatID+2*n)] -= phir*(LMgam/(s1*d1)+LPgam/(r*d1))-phil*RMgam/(r*d2);
        
        A[2*n+xMatID+nOFm*(xMatID+2*np1)]   -= phir*(RMgam/(r*d2)+RPgam/(s3*d2))-phil*LPgam/(r*d1);
        A[2*np1+xMatID+nOFm*(xMatID+2*np1)] -= phil*(RMgam/(r*d2)+RPgam/(s3*d2))-phir*LPgam/(r*d1);
        
        A[2*n+xMatID+nOFm*(2*np2+xMatID)]   += RPgam*phir/(s3*d2);
        A[2*np1+xMatID+nOFm*(2*np2+xMatID)] += RPgam*phil/(s3*d2);
        
        A[xMatID+2*n+1+nOFm*(xMatID+2*nm1+1)]   += LMgam*phil/(s1*d1);
        A[xMatID+2*np1+1+nOFm*(xMatID+2*nm1+1)] += LMgam*phir/(s1*d1);
        
        A[(xMatID+2*n+1)*(nOFm+1)]            -= phil*(LMgam/(s1*d1)+LPgam/(r*d1))-phir*RMgam/(r*d2);
        A[xMatID+2*np1+1+nOFm*(xMatID+2*n+1)] -= phir*(LMgam/(s1*d1)+LPgam/(r*d1))-phil*RMgam/(r*d2);
        
        A[xMatID+2*n+1+nOFm*(xMatID+2*np1+1)]   -= phir*(RMgam/(r*d2)+RPgam/(s3*d2))-phil*LPgam/(r*d1);
        A[xMatID+2*np1+1+nOFm*(xMatID+2*np1+1)] -= phil*(RMgam/(r*d2)+RPgam/(s3*d2))-phir*LPgam/(r*d1);
        
        A[xMatID+2*n+1+nOFm*(2*np2+xMatID+1)]   += phir*RPgam/(s3*d2);
        A[xMatID+2*np1+1+nOFm*(2*np2+xMatID+1)] += phil*RPgam/(s3*d2);
		
		//	Loop over all panels, switch for singularity position (midpoint corner)
		for(int k=0; k<ifblock; k++){
			rmtElms = bem->Elements[k].getFullSize();
            Xs = bem->Elements[k].XId[0];
            Ys = bem->Elements[k].YId[0];	
            #pragma omp parallel
			{
				#pragma omp for private(xgs,ygs,x1,y1,G11,G12,G22,T111,T112,T122,T211,T212,T222,phil,phir,dolog,fixind)
				//	Loop over all elements of a panel
				for(int m=0; m<rmtElms; m++){
					xgs = xls-0.5*(Xs[m+1]+Xs[m]);
                    ygs = yls-0.5*(Ys[m+1]+Ys[m]);
                    fixind = yMatID+nOFm*xMatID+2*m;
                    
					// Loop over Gauss points
					for(int l=0; l<6; l++){
						x1 = 0.5*dx*IGP[l]+xgs;
						y1 = 0.5*dy*IGP[l]+ygs; 
					
						computeGT(1, x1, y1, G11, G12, G22, T111, T112, T122, T211, T212, T222);
					
						phil = (1.0-IGP[l])*0.25*IGW[l]*r;
						phir = (1.0+IGP[l])*0.25*IGW[l]*r;
						
						rhs[yMatID+2*m]   += (fl1*phil+fr1*phir)*G11+(fl2*phil+fr2*phir)*G12;
						rhs[yMatID+2*m+1] += (fl1*phil+fr1*phir)*G12+(fl2*phil+fr2*phir)*G22;
						
						A[fixind+2*n*nOFm] += dLambda*phil*(T111*lnx+T112*lny);
						A[fixind+(2*n+1)*nOFm] += dLambda*phil*(T112*lnx+T122*lny);
						A[fixind+nOFm*2*np1] += dLambda*phir*(T111*lnx+T112*lny);
						A[fixind+nOFm*(2*np1+1)] += dLambda*phir*(T112*lnx+T122*lny);
						
						A[fixind+2*n*nOFm+1] += dLambda*phil*(T211*lnx+T212*lny);
						A[fixind+2*n*nOFm+1+nOFm] += dLambda*phil*(T212*lnx+T222*lny);
						A[fixind+1+nOFm*(2*np1)]  += dLambda*phir*(T211*lnx+T212*lny);
						A[fixind+1+nOFm*(2*np1+1)]+= dLambda*phir*(T212*lnx+T222*lny);
						
						// Implicit Interface
						
						phil = delT*phil;
						phir = delT*phir;
						
                        // T111 etc. variables are reused because they are not used before being reset and they are omp private
                        
                        T111 = LMgam*phil/(s1*d1);
                        T222 = RPgam*phir/(s3*d2);
                        T112 = phil*(LMgam/(s1*d1)+LPgam/(r*d1))-phir*RMgam/(r*d2);
                        T211 = phir*(RMgam/(r*d2)+RPgam/(s3*d2))-phil*LPgam/(r*d1);
                        
                        // Test function x-direction
						A[fixind+2*nm1*nOFm]	  += -G11*T111;  // ull
						A[fixind+(2*nm1+1)*nOFm] += -G12*T111;    // vll
                        
                        A[fixind+2*n*nOFm]	   += G11*T112; //ul
						A[fixind+(2*n+1)*nOFm] += G12*T112; //vl
						
                        A[fixind+2*np1*nOFm]	  += G11*T211; // ur
						A[fixind+(2*np1+1)*nOFm]  += G12*T211; // vr
						
                        A[fixind+2*np2*nOFm]	 += -G11*T222; // urr
						A[fixind+(2*np2+1)*nOFm] += -G12*T222; // vrr
						 
                        // Test function y-direction
						A[fixind+1+2*nm1*nOFm]     += -G12*T111; // ull
						A[fixind+1+(2*nm1+1)*nOFm] += -G22*T111; // vll
						
                        A[fixind+1+2*n*nOFm]		+= G12*T112; // ul
						A[fixind+1+(2*n+1)*nOFm]	+= G22*T112; // vl
						
                        A[fixind+1+2*np1*nOFm]     += G12*T211; // ur
						A[fixind+1+(2*np1+1)*nOFm] += G22*T211; // vr
						
                        A[fixind+1+2*np2*nOFm]	    += -G12*T222;   // urr
						A[fixind+1+(2*np2+1)*nOFm]	+= -G22*T222;   // vrr
					}
				}
			}
            # pragma omp barrier
			yMatID += 2*rmtElms;
		}
		
		for(int k=ifblock; k<nbBlocks; k++){
			rmtElms = bem->Elements[k].getFullSize();
            Xs = bem->Elements[k].XId[0];
            Ys = bem->Elements[k].YId[0];
            fixind = yMatID*nOFm+ xMatID+2*n;
			#pragma omp parallel
			{
				#pragma omp for private(xgs,ygs,x1,y1,G11,G12,G22,T111,T112,T122,T211,T212,T222,phil,phir,dolog,fixind)
				//	Loop over all elements of a panel
				for(int m=0; m<rmtElms; m++){
					xgs = xls-Xs[m];
					ygs = yls-Ys[m];
					fixind = yMatID+nOFm*xMatID+2*m;
                    
					// Decide wheather it is a singular element
                    dolog = (!(((n==m)||(np1==m))&&(k==(inP-bem->ifid+ifblock))));
					
					// Loop over Gauss points
					for(int l=0; l<6; l++){
						x1 = 0.5*dx*IGP[l]+xgs;
						y1 = 0.5*dy*IGP[l]+ygs; 
						
						computeGT(dolog, x1, y1, G11, G12, G22, T111, T112, T122, T211, T212, T222);
						
						phil = (1.0-IGP[l])*0.25*IGW[l]*r;
						phir = (1.0+IGP[l])*0.25*IGW[l]*r;
						
						rhs[yMatID+2*m]   += (fl1*phil+fr1*phir)*G11+(fl2*phil+fr2*phir)*G12;
						rhs[yMatID+2*m+1] += (fl1*phil+fr1*phir)*G12+(fl2*phil+fr2*phir)*G22;
						
						A[fixind+2*n*nOFm]  += dLambda*phil*(T111*lnx+T112*lny);
						A[fixind+(2*n+1)*nOFm] += dLambda*phil*(T112*lnx+T122*lny);
						A[fixind+nOFm*2*np1]	   += dLambda*phir*(T111*lnx+T112*lny);
						A[fixind+nOFm*(2*np1+1)]  += dLambda*phir*(T112*lnx+T122*lny);
						
						A[fixind+1+2*n*nOFm] += dLambda*phil*(T211*lnx+T212*lny);
						A[fixind+1+(2*n+1)*nOFm] += dLambda*phil*(T212*lnx+T222*lny);
						A[fixind+1+nOFm*(2*np1)]  += dLambda*phir*(T211*lnx+T212*lny);
						A[fixind+1+nOFm*(2*np1+1)]+= dLambda*phir*(T212*lnx+T222*lny);
						
						// Implicit Interface

						phil = delT*phil;
						phir = delT*phir;
                        
                        A[fixind+2*nm1*nOFm]	  += -LMgam*G11*phil/(s1*d1);  // ull
						A[fixind+(2*nm1+1)*nOFm] += -LMgam*G12*phil/(s1*d1);    // vll
						
                        A[fixind+2*n*nOFm]	   += G11*(phil*(LMgam/(s1*d1)+LPgam/(r*d1))-phir*RMgam/(r*d2)); //ul
						A[fixind+(2*n+1)*nOFm] += G12*(phil*(LMgam/(s1*d1)+LPgam/(r*d1))-phir*RMgam/(r*d2));
						
                        A[fixind+2*np1*nOFm]	  += G11*(phir*(RMgam/(r*d2)+RPgam/(s3*d2))-phil*LPgam/(r*d1)); // ur
						A[fixind+(2*np1+1)*nOFm]  += G12*(phir*(RMgam/(r*d2)+RPgam/(s3*d2))-phil*LPgam/(r*d1));
						
                        A[fixind+2*np2*nOFm]	 += -RPgam*G11*phir/(s3*d2); // urr
						A[fixind+(2*np2+1)*nOFm] += -RPgam*G12*phir/(s3*d2);
                        
                        // Test function y-direction
						A[fixind+1+2*nm1*nOFm]     += -LMgam*G12*phil/(s1*d1);
						A[fixind+1+(2*nm1+1)*nOFm] += -LMgam*G22*phil/(s1*d1);
						
                        A[fixind+1+2*n*nOFm]		+= G12*(phil*(LMgam/(s1*d1)+LPgam/(r*d1))-phir*RMgam/(r*d2));
						A[fixind+1+(2*n+1)*nOFm]	+= G22*(phil*(LMgam/(s1*d1)+LPgam/(r*d1))-phir*RMgam/(r*d2));
						
                        A[fixind+1+2*np1*nOFm]     += G12*(phir*(RMgam/(r*d2)+RPgam/(s3*d2))-phil*LPgam/(r*d1));
						A[fixind+1+(2*np1+1)*nOFm] += G22*(phir*(RMgam/(r*d2)+RPgam/(s3*d2))-phil*LPgam/(r*d1));
						
                        A[fixind+1+2*np2*nOFm]	    += -RPgam*G12*phir/(s3*d2);
						A[fixind+1+(2*np2+1)*nOFm]	+= -RPgam*G22*phir/(s3*d2);

					}
				}
			}
            # pragma omp barrier
			yMatID += 2*rmtElms;
		}
	}
    
    dominantForce = std::max(maxForce-minForce,dominantForce);
}


// void MatBlC1::FilmTrac(double *erhs){

// void MatBlC1::BordTrac(int inP,double delT){

void MatBlC1::BoosThess(int inP){
    // Boundary contion to test Boos and Thess
	double dx, dy, xls, yls, xgs, ygs, r, lnx, lny, f;
//	double cx, cy, cr, cri, cre;
	double x1, y1, G11, G12, G22;
	double T111, T122, T112, T211, T222, T212, gvar;
    double *Xp, *Yp;  // Xp for panels and Xs for singularities
	int localElm = bem->allSize[inP];
	double dLambda = Lambda -1.0;
	
	int np1, dolog;
	
    Xp = bem->Elements[0].XId[0];
    Yp = bem->Elements[0].YId[0];
    
	for(int n=0; n<localElm; n++){
		if (n!=localElm-1) {
			np1 = n+1;			// np points to the next base function
		} else {
			np1 = 0;
		}
		
		dx = Xp[n+1]-Xp[n];					
        dy = Yp[n+1]-Yp[n];
        xls = 0.5*(Xp[n+1]+Xp[n]);
        yls = 0.5*(Yp[n+1]+Yp[n]);
        r = (dx*dx+dy*dy);
        r = sqrt(r);
        lnx = dy/r;
        lny = -dx/r;
        f = lnx; //bem->Elements[0].BCvalueX[0]*lnx;  // it shall be x! CHECK
        
     /*   cx = Xp[n]-Xp[n-1];
		cy = Yp[n]-Yp[n-1];
        
        cr = sqrt(cx*cx+cy*cy);
        cr = (dx*cx+dy*cy)/(r*cr);
        
        cri = -2*(M_PI-acos(cr));
        cre = -2*(M_PI+acos(cr));*/
        
  //      A[2*n*(nOFm+1)]	  += lny*(cri*Eta+cre*(1-Eta));
  //      A[2*n*(nOFm+1)+1] += -lnx*(cri*Eta+cre*(1-Eta));
    
        
        // 0th order
        A[2*n*(nOFm+1)]	  += lny*2.0*M_PI*(Lambda+1.0);
        A[2*n*(nOFm+1)+1] += -lnx*2.0*M_PI*(Lambda+1.0);
     
        A[(2*n+1)*(nOFm+1)-1] += -r*(log(r/2)-1)*lnx;
        A[(2*n+1)*(nOFm+1)] += -r*(log(r/2)-1)*lny;
        
        rhs[2*n]   += -f*r*(log(r/2)-1)*lny;
        rhs[2*n+1] += f*r*(log(r/2)-1)*lnx;
        
		for(int m=0; m<localElm; m++){
		//	xgs = xls-Xp[m];
		//	ygs = yls-Yp[m];
			
            // zeroth order
            xgs = xls-0.5*(Xp[m+1]+Xp[m]);
            ygs = yls-0.5*(Yp[m+1]+Yp[m]);
            
            dolog = !(n==m);
            
			for(int l=0; l<6; l++){
				x1 = 0.5*dx*IGP[l]+xgs;
				y1 = 0.5*dy*IGP[l]+ygs; 
				
				gvar = IGW[l]*r*0.5;
                
				computeGT(dolog, x1, y1, G11, G12, G22, T111, T112, T122, T211, T212, T222);
				
                
                
	//			A[2*m+2*n*nOFm]	    += dEta*(1.0-IGP[l])*(-lny*(T111*lnx+T112*lny)+lnx*(T112*lnx+T122*lny))*gvar*0.5;
				A[2*m+(2*n+1)*nOFm]   += -(G11*lnx+G12*lny)*gvar;
	//			A[2*m+2*np1*nOFm]	+= dEta*(1.0+IGP[l])*(-lny*(T111*lnx+T112*lny)+lnx*(T112*lnx+T122*lny))*gvar*0.5;
					
	//			A[2*m+1+2*n*nOFm]   += dEta*(1.0-IGP[l])*(-lny*(T211*lnx+T212*lny)+lnx*(T212*lnx+T222*lny))*gvar*0.5;
				A[2*m+1+(2*n+1)*nOFm] += -(G12*lnx+G22*lny)*gvar;
		//		A[2*m+1+2*np1*nOFm] += dEta*(1.0+IGP[l])*(-lny*(T211*lnx+T212*lny)+lnx*(T212*lnx+T222*lny))*gvar*0.5;
				
                // zeroth order
                A[2*m+2*n*nOFm]	    +=  dLambda*(-lny*(T111*lnx+T112*lny)+lnx*(T112*lnx+T122*lny))*gvar;
                A[2*m+1+2*n*nOFm] += dLambda*(-lny*(T211*lnx+T212*lny)+lnx*(T212*lnx+T222*lny))*gvar;

                
				rhs[2*m]   += (-G11*lny+G12*lnx)*f*gvar;
				rhs[2*m+1] += (-G12*lny+G22*lnx)*f*gvar;
			}
		}
	}
}

double MatBlC1::getUvel(int dropId, int nm1, int n, int np1){
	// THIS HAS BUG, the addressing can be wrong
	double U,V,Nx,Ny,R;
	int selvel =0;
	for (int k=0; k<dropId; k++) {
		selvel += 2*bem->allSize[k];
	}

	U = rhs[selvel+2*n];
	V = rhs[selvel+2*n+1];
	Ny = bem->allXId[dropId][np1]-bem->allXId[dropId][nm1];
	Nx = -(bem->allYId[dropId][np1]-bem->allYId[dropId][nm1]);
	R = sqrt(Nx*Nx+Ny*Ny);
	Nx = Nx/R;
	Ny = Ny/R;
	return (U*Nx+V*Ny);
	//  return 0.0;
}

void MatBlC1::getForce(int WallId, int nbSeg, double* Force, double* sol){
	// calculates the forces on a wall segment
	double* Xc;
	double* Yc;
	double F1=0;
	double F2=0;
	double F3 = 0;
	double dx,dy, ax, ay,rr, xcm, ycm;
	int spacer = 0;

	bem->getCoM(bem->panel2block[WallId], xcm, ycm);

	for (int k=WallId; k<(WallId+nbSeg); k++) {
		Xc = bem->allXId[k];
		Yc = bem->allYId[k];
		for (int n=0; n<bem->allSize[k]; n++) {
			dx = Xc[n+1]-Xc[n];
			dy = Yc[n+1]-Yc[n];
			ax = 0.5*(Xc[n+1]+Xc[n])-xcm;
			ay = 0.5*(Yc[n+1]+Yc[n])-ycm;
			rr = sqrt(dx*dx+dy*dy);

			//    F1 += -sol[2*n+spacer]*dy+sol[2*n+1+spacer]*dx;
			//    F2 += sol[2*n+spacer]*dx+sol[2*n+1+spacer]*dy;
			F1 += sol[2*n+spacer]*rr;
			F2 += sol[2*n+spacer+1]*rr;
			F3 += -sol[2*n+spacer]*rr*ay+sol[2*n+spacer+1]*rr*ax;
		}
		spacer += 2*bem->allSize[k];
	}

	Force[0] = F1;
	Force[1] = F2;
	Force[2] = F3;
}

// void MatBlC1::ForceFree(int fblock, double dEta, double dgap){

// -------------------------- Dedicated to Accelerated Boundary conditions that are used when Matrix is Precondensed


void MatBlC1::SZeroVel(int inP){

	/*	Integration from all singularities over a boundary with zero velocity.
		Wall Elements are always zeroth order
	 */

	double dx, dy, xls, yls, xgs, ygs, r;
	double x1, y1, gvar, G11, G12, G22;
	double *Xp, *Yp, *Xs, *Ys;  // Xp for panels and Xs for singularities
	int localElm = bem->allSize[inP];
	int rmtElms, fixind, mp1;
	int xMatID = 4*fixsize*fixsize;
	int yMatID = 0;
	int NOFN = nOFm-2*fixsize;

	Xp = bem->allXId[inP];
	Yp = bem->allYId[inP];

	for(int m=0; m<inP; m++){
		xMatID += 2*NOFN*bem->allSize[m];
	}

	//	Size of the Element, the Position of its center, Normals
	for(int n=0; n<localElm; n++){
		dx = Xp[n+1]-Xp[n];
		dy = Yp[n+1]-Yp[n];
		xls = 0.5*(Xp[n+1]+Xp[n]);
		yls = 0.5*(Yp[n+1]+Yp[n]);
		r = (dx*dx+dy*dy);
		r = sqrt(r);

		yMatID = 0;

		//	Loop over all panels, switch for singularity position (midpoint corner)
		for(int k=dynblock; k<nbBlocks; k++){
			rmtElms = bem->Elements[k].getFullSize();
			Xs = bem->Elements[k].XId[0];
			Ys = bem->Elements[k].YId[0];
			fixind = yMatID+xMatID+2*n*NOFN;
			mp1 = (k<bem->ifblock);
#pragma omp parallel
			{

#pragma omp for private(xgs,ygs,x1,y1,G11,G12,G22,gvar)
				//	Loop over all elements of a panel
				for(int m=0; m<rmtElms; m++){

					xgs = xls-0.5*(Xs[m+mp1]+Xs[m]);
					ygs = yls-0.5*(Ys[m+mp1]+Ys[m]);

					gvar = xgs*xgs+ygs*ygs;

					// Decide wheather it is a singular element

					// Loop over Gauss points
					for(int l=0; l<6; l++){
						x1 = 0.5*dx*IGP[l]+xgs;
						y1 = 0.5*dy*IGP[l]+ygs;

						computeGO(1, x1, y1, G11, G12, G22);
						gvar = IGW[l]*r*0.5;

						A[fixind+2*m] -= G11*gvar;
						A[fixind+NOFN+2*m] -= G12*gvar;
						A[fixind+1+2*m]  -= G12*gvar;
						A[fixind+NOFN+1+2*m] -= G22*gvar;
					}
				}
			}
# pragma omp barrier
			yMatID += 2*rmtElms;
		}
	}
}


void MatBlC1::SFlatVel(int inP, double Vn, double Vt, double Theta, double dEta){

	/*	Integration from all singularities over a boundary with n-t velocity.
		Wall Elements are always zeroth order
	 */

	double dx, dy, xls, yls, xgs, ygs, gvar, r, lnx, lny, u1, u2, xcm, ycm;
	double x1, y1, G11, G12, G22, T111, T122, T112, T211, T222, T212;

	double *Xp, *Yp, *Xs, *Ys;  // Xp for panels and Xs for singularities
	int localElm = bem->allSize[inP];
	int rmtElms, fixind, dolog, mp1;

	int NOFN = nOFm-2*fixsize;
	int MOFM = 2*fixsize;
	int xMatID1 = 0;
	int ind1 = 2*fixsize*(2*fixsize+NOFN);
	int xMatID2 = 0;
	int ind2 = 2*fixsize*(2*fixsize+2*NOFN);
	int yMatID = 0;

	Xp = bem->allXId[inP];
	Yp = bem->allYId[inP];

	bem->getCoM(bem->panel2block[inP], xcm, ycm);

	for(int m=dynpanel; m<inP; m++){
		xMatID1 += 2*fixsize*bem->allSize[m];
		xMatID2 += 2*bem->allSize[m];
	}

	//	Size of the Element, the Position of its center, normals
	for(int n=0; n<localElm; n++){
		dx = Xp[n+1]-Xp[n];
		dy = Yp[n+1]-Yp[n];
		xls = 0.5*(Xp[n+1]+Xp[n]);
		yls = 0.5*(Yp[n+1]+Yp[n]);
		r = (dx*dx+dy*dy);
		r = sqrt(r);
		lnx = dy/r;
		lny = -dx/r;
		u1 = Vn - (yls-ycm)*Theta;
		u2 = Vt + (xls-xcm)*Theta;
		yMatID = 0;
		mp1 = 1;

		A[ind2+(2*n+xMatID2)*(NOFN+1)] -= r*(log(r/2)-1);
		A[ind2+(2*n+1+xMatID2)*(NOFN+1)] -= r*(log(r/2)-1);

		rhs[2*fixsize+xMatID2+2*n] += 2*dEta*M_PI*u1;
		rhs[2*fixsize+ xMatID2+2*n+1] += 2*dEta*M_PI*u2;

		//	Loop over all panels, switch for singularity position (midpoint corner)
		for(int k=0; k<dynblock; k++){
			rmtElms = bem->Elements[k].getFullSize();
			Xs = bem->Elements[k].XId[0];
			Ys = bem->Elements[k].YId[0];
			fixind = ind1 + yMatID+xMatID1;
			mp1 = (k<bem->ifblock);
#pragma omp parallel
			{

#pragma omp for private(xgs,ygs,x1,y1,G11,G12,G22,T111,T112,T122,T211,T212,T222,gvar)
				//	Loop over all elements of a panel
				for(int m=0; m<rmtElms; m++){
					xgs = xls-0.5*(Xs[m+mp1]+Xs[m]);
					ygs = yls-0.5*(Ys[m+mp1]+Ys[m]);

					gvar = xgs*xgs+ygs*ygs;

					// Loop over Gauss points
					for(int l=0; l<6; l++){
						x1 = 0.5*dx*IGP[l]+xgs;
						y1 = 0.5*dy*IGP[l]+ygs;

						computeGT(1, x1, y1, G11, G12, G22, T111, T112, T122, T211, T212, T222);
						gvar = IGW[l]*r*0.5;

						A[fixind+2*n*MOFM+2*m] -= G11*gvar;
						A[fixind+(2*n+1)*MOFM+2*m] -= G12*gvar;
						A[fixind+MOFM*2*n+1+2*m] -= G12*gvar;
						A[fixind+MOFM*(2*n+1)+1+2*m] -= G22*gvar;

						rhs[yMatID+2*m]   -= dEta*((T111*lnx+T112*lny)*u1+(T112*lnx+T122*lny)*u2)*gvar;
						rhs[yMatID+2*m+1] -= dEta*((T211*lnx+T212*lny)*u1+(T212*lnx+T222*lny)*u2)*gvar;
					}
				}
			}
# pragma omp barrier
			yMatID += 2*rmtElms;
		}

		yMatID = 0;

		for(int k=dynblock; k<nbBlocks; k++){
			rmtElms = bem->Elements[k].getFullSize();
			Xs = bem->Elements[k].XId[0];
			Ys = bem->Elements[k].YId[0];
			fixind = ind2+yMatID+ xMatID2*NOFN;
			mp1 = (k<bem->ifblock);
#pragma omp parallel
			{

#pragma omp for private(xgs,ygs,x1,y1,G11,G12,G22,T111,T112,T122,T211,T212,T222,gvar,dolog)
				//	Loop over all elements of a panel
				for(int m=0; m<rmtElms; m++){
					xgs = xls-0.5*(Xs[m+mp1]+Xs[m]);
					ygs = yls-0.5*(Ys[m+mp1]+Ys[m]);

					gvar = xgs*xgs+ygs*ygs;
					dolog = (gvar>1e-12);

					// Loop over Gauss points
					for(int l=0; l<6; l++){
						x1 = 0.5*dx*IGP[l]+xgs;
						y1 = 0.5*dy*IGP[l]+ygs;

						computeGT(dolog, x1, y1, G11, G12, G22, T111, T112, T122, T211, T212, T222);
						gvar = IGW[l]*r*0.5;

						A[fixind+2*n*NOFN+2*m] -= G11*gvar;
						A[fixind+(2*n+1)*NOFN+2*m] -= G12*gvar;
						A[fixind+NOFN*2*n+1+2*m] -= G12*gvar;
						A[fixind+NOFN*(2*n+1)+1+2*m] -= G22*gvar;

						rhs[yMatID+2*m]   -= dEta*((T111*lnx+T112*lny)*u1+(T112*lnx+T122*lny)*u2)*gvar;
						rhs[yMatID+2*m+1] -= dEta*((T211*lnx+T212*lny)*u1+(T212*lnx+T222*lny)*u2)*gvar;
					}
				}
			}
# pragma omp barrier
			yMatID += 2*rmtElms;
		}
	}
	//    std::cout<<A[6727378]<<"\n";
}

// void MatBlC1::SFlatVelrhs(int inP, double Vn, double Vt, double Theta, double dEta, int insert){

void MatBlC1::SNormVel(int inP, double Vn, double Vt, double dEta){

	/*	Integration from all singularities over a boundary with n-t velocity.
		Wall Elements are always zeroth order
	 */

	double dx, dy, xls, yls, xgs, ygs, gvar, r, lnx, lny, u1, u2;
	double x1, y1, G11, G12, G22, T111, T122, T112, T211, T222, T212;

	double *Xp, *Yp, *Xs, *Ys;  // Xp for panels and Xs for singularities
	int localElm = bem->allSize[inP];
	int rmtElms, fixind, mp1;
	int xMatID = 4*fixsize*fixsize;
	int yMatID = 0;
	int NOFN = nOFm-2*fixsize;

	Xp = bem->allXId[inP];
	Yp = bem->allYId[inP];

	for(int m=0; m<inP; m++){
		xMatID += 2*NOFN*bem->allSize[m];
	}

	//	Size of the Element, the Position of its center, normals
	for(int n=0; n<localElm; n++){
		dx = Xp[n+1]-Xp[n];
		dy = Yp[n+1]-Yp[n];
		xls = 0.5*(Xp[n+1]+Xp[n]);
		yls = 0.5*(Yp[n+1]+Yp[n]);
		r = (dx*dx+dy*dy);
		r = sqrt(r);
		lnx = dy/r;
		lny = -dx/r;
		u1 = Vn*lnx-Vt*lny;
		u2 = Vn*lny+Vt*lnx;
		yMatID = 0;

		for(int k=dynblock; k<nbBlocks; k++){
			rmtElms = bem->Elements[k].getFullSize();
			Xs = bem->Elements[k].XId[0];
			Ys = bem->Elements[k].YId[0];
			fixind = yMatID+xMatID+2*n*NOFN;
			mp1 = (k<bem->ifblock);
#pragma omp parallel
			{

#pragma omp for private(xgs,ygs,x1,y1,G11,G12,G22,T111,T112,T122,T211,T212,T222,gvar,fixind)
				//	Loop over all elements of a panel
				for(int m=0; m<rmtElms; m++){

					xgs = xls-0.5*(Xs[m+mp1]+Xs[m]);
					ygs = yls-0.5*(Ys[m+mp1]+Ys[m]);

					gvar = xgs*xgs+ygs*ygs;
					// Loop over Gauss points
					for(int l=0; l<6; l++){
						x1 = 0.5*dx*IGP[l]+xgs;
						y1 = 0.5*dy*IGP[l]+ygs;

						computeGT(1, x1, y1, G11, G12, G22, T111, T112, T122, T211, T212, T222);
						gvar = IGW[l]*r*0.5;

						A[fixind+2*m] -= G11*gvar;
						A[fixind+NOFN+2*m] -= G12*gvar;
						A[fixind+1+2*m] -= G12*gvar;
						A[fixind+NOFN+1+2*m] -= G22*gvar;

						rhs[yMatID+2*m]   -= dEta*((T111*lnx+T112*lny)*u1+(T112*lnx+T122*lny)*u2)*gvar;
						rhs[yMatID+2*m+1] -= dEta*((T211*lnx+T212*lny)*u1+(T212*lnx+T222*lny)*u2)*gvar;
					}
				}
			}
# pragma omp barrier
			yMatID += 2*rmtElms;
		}
	}
}

void MatBlC1::SPOX(int inP, double plevel, double dEta){
	/*	Integration from all singularities over a boundary pressure.
		Wall Elements are always zeroth order
	 */

	double dx, dy, xls, yls, xgs, ygs, r, lnx, lny, f1;
	double x1, y1, G11, G12, G22, gvar;
	double T111, T122, T112, T211, T222, T212;
	double *Xp, *Yp, *Xs, *Ys;  // Xp for panels and Xs for singularities
	int localElm = bem->allSize[inP];
	int fixind;

	int rmtElms, mp1;
	int xMatID = 4*fixsize*fixsize;
	int yMatID = 0;
	int NOFN = nOFm-2*fixsize;
	Xp = bem->allXId[inP];
	Yp = bem->allYId[inP];

	for(int m=0; m<inP; m++){
		xMatID += 2*NOFN*bem->allSize[m];
	}

	//	Size of the Element, the Position of its center, Normals
	for(int n=0; n<localElm; n++){
		dx = Xp[n+1]-Xp[n];
		dy = Yp[n+1]-Yp[n];
		xls = 0.5*(Xp[n+1]+Xp[n]);
		yls = 0.5*(Yp[n+1]+Yp[n]);
		r = dx*dx+dy*dy;
		r = sqrt(r);
		lnx = dy/r;
		lny = -dx/r;
		f1 = -plevel*lnx;
		yMatID = 0;

		for(int k=dynblock; k<nbBlocks; k++){
			rmtElms = bem->Elements[k].getFullSize();
			Xs = bem->Elements[k].XId[0];
			Ys = bem->Elements[k].YId[0];
			fixind = yMatID+xMatID+2*n*NOFN;
			mp1 = (k<bem->ifblock);
#pragma omp parallel
			{
#pragma omp for private(xgs,ygs,x1,y1,G11,G12,G22,T111,T112,T122,T211,T212,T222,gvar)
				//	Loop over all elements of a panel
				for(int m=0; m<rmtElms; m++){

					xgs = xls-0.5*(Xs[m+mp1]+Xs[m]);
					ygs = yls-0.5*(Ys[m+mp1]+Ys[m]);

					gvar = xgs*xgs+ygs*ygs;
					// Loop over Gauss points
					for(int l=0; l<6; l++){
						x1 = 0.5*dx*IGP[l]+xgs;
						y1 = 0.5*dy*IGP[l]+ygs;

						computeGT(1,x1, y1, G11, G12, G22, T111, T112, T122, T211, T212, T222);
						gvar = IGW[l]*r*0.5;

						A[fixind+2*m] += dEta*(T111*lnx+T112*lny)*gvar;
						A[fixind+NOFN+2*m] -= G12*gvar;
						A[fixind+1+2*m] += dEta*(T211*lnx+T212*lny)*gvar;
						A[fixind+NOFN+1+2*m] -= G22*gvar;

						rhs[yMatID+2*m+2*fixsize]   += (f1*G11)*gvar;
						rhs[yMatID+2*m+1+2*fixsize] += (f1*G12)*gvar;
					}
				}
			}
# pragma omp barrier
			yMatID += 2*rmtElms;
		}
	}
}

void MatBlC1::SPressure(int inP, double plevel, double dEta){

	/*	Integration from all singularities over a boundary pressure.
		Wall Elements are always zeroth order
	 */

	double dx, dy, xls, yls, xgs, ygs, r, lnx, lny, f1, f2;
	double x1, y1, G11, G12, G22, gvar;
	double T111, T122, T112, T211, T222, T212;
	double *Xp, *Yp, *Xs, *Ys;  // Xp for panels and Xs for singularities
	int localElm = bem->allSize[inP];

	int rmtElms, fixind, mp1;
	int xMatID = 4*fixsize*fixsize;
	int yMatID = 0;
	int NOFN = nOFm-2*fixsize;

	Xp = bem->allXId[inP];
	Yp = bem->allYId[inP];

	for(int m=0; m<inP; m++){
		xMatID += 2*NOFN*bem->allSize[m];
	}

	//	Size of the Element, the Position of its center, Normals
	for(int n=0; n<localElm; n++){
		dx = Xp[n+1]-Xp[n];
		dy = Yp[n+1]-Yp[n];
		xls = 0.5*(Xp[n+1]+Xp[n]);
		yls = 0.5*(Yp[n+1]+Yp[n]);
		r = dx*dx+dy*dy;
		r = sqrt(r);
		lnx = dy/r;
		lny = -dx/r;
		f1 = plevel*lnx;
		f2 = plevel*lny;
		yMatID = 0;

		//	Loop over all panels, switch for singularity position (midpoint corner)
		for(int k=dynblock; k<nbBlocks; k++){
			rmtElms = bem->Elements[k].getFullSize();
			Xs = bem->Elements[k].XId[0];
			Ys = bem->Elements[k].YId[0];
			fixind = yMatID+xMatID+2*n*NOFN;
			mp1 = (k<bem->ifblock);
#pragma omp parallel
			{

#pragma omp for private(xgs,ygs,x1,y1,G11,G12,G22,T111,T112,T122,T211,T212,T222,gvar)
				//	Loop over all elements of a panel
				for(int m=0; m<rmtElms; m++){

					xgs = xls-0.5*(Xs[m+mp1]+Xs[m]);
					ygs = yls-0.5*(Ys[m+mp1]+Ys[m]);

					gvar = xgs*xgs+ygs*ygs;

					// Loop over Gauss points
					for(int l=0; l<6; l++){
						x1 = 0.5*dx*IGP[l]+xgs;
						y1 = 0.5*dy*IGP[l]+ygs;

						computeGT(1, x1, y1, G11, G12, G22, T111, T112, T122, T211, T212, T222);
						gvar = IGW[l]*r*0.5;

						A[fixind+2*m] += dEta*(T111*lnx+T112*lny)*gvar;
						A[fixind+NOFN+2*m] += dEta*(T112*lnx+T122*lny)*gvar;
						A[fixind+1+2*m] += dEta*(T211*lnx+T212*lny)*gvar;
						A[fixind+NOFN+1+2*m] += dEta*(T212*lnx+T222*lny)*gvar;

						rhs[yMatID+2*m+2*fixsize]   += (f1*G11+f2*G12)*gvar;
						rhs[yMatID+2*m+1+2*fixsize] += (f1*G12+f2*G22)*gvar;

					}
				}
			}
# pragma omp barrier
			yMatID += 2*rmtElms;
		}
	}

}

void MatBlC1::SPoiseuille(int inP, double qdebit, double dEta){
	double dx, dy, xls, yls, xgs, ygs, r, lnx, lny, u1, u2;
	double x1, y1, gvar, G11, G12, G22;
	double T111, T122, T112, T211, T222, T212;
	double dxpanel, dypanel, xppanel, yppanel, vel, Vn, dpanel, coshb, pmr;
	double *Xp, *Yp, *Xs, *Ys;  // Xp for panels and Xs for singularities
	int localElm = bem->allSize[inP];
	int NOFN = nOFm-2*fixsize;
	int rmtElms, fixind, mp1;
	int xMatID = 4*fixsize*fixsize;
	int yMatID = 0;

	Xp = bem->allXId[inP];
	Yp = bem->allYId[inP];

	for(int m=0; m<inP; m++){
		xMatID += 2*NOFN*bem->allSize[m];
	}

	// Determine Panel Parameters
	dypanel = Yp[localElm]-Yp[0];
	dxpanel = Xp[localElm]-Xp[0];
	xppanel = 0.5*(Xp[localElm]+Xp[0]);
	yppanel = 0.5*(Yp[localElm]+Yp[0]);
	dpanel = sqrt(dxpanel*dxpanel+dypanel*dypanel);
	//	nxpanel = dypanel/dpanel;
	//	nypanel = -dxpanel/dpanel;
	coshb = cosh(kF*dpanel*0.5);
	vel = qdebit*dpanel*kF/(dpanel*kF*coshb-2*sinh(kF*dpanel/2));

	//	Size of the Element, the Position of its center, normals
	for(int n=0; n<localElm; n++){
		dx = Xp[n+1]-Xp[n];
		dy = Yp[n+1]-Yp[n];
		xls = 0.5*(Xp[n+1]+Xp[n]);
		yls = 0.5*(Yp[n+1]+Yp[n]);
		r = (dx*dx+dy*dy);
		r = sqrt(r);
		lnx = dy/r;
		lny = -dx/r;
		pmr = sqrt((xls-xppanel)*(xls-xppanel)+(yls-yppanel)*(yls-yppanel));
		if(kF!=0){
			Vn = vel*(coshb-cosh(kF*pmr));
		} else {
			Vn = qdebit/dpanel*1.5*(1.0-4*(pmr/dpanel)*(pmr/dpanel));
		}
		u1 = -Vn*lnx;
		u2 = -Vn*lny;
		yMatID = 0;

		//	Loop over all panels, switch for singularity position (midpoint corner)
		for(int k=dynblock; k<nbBlocks; k++){
			rmtElms = bem->Elements[k].getFullSize();
			Xs = bem->Elements[k].XId[0];
			Ys = bem->Elements[k].YId[0];
			fixind = yMatID+xMatID+2*n*NOFN;
			mp1 = (k<bem->ifblock);
#pragma omp parallel
			{

#pragma omp for private(xgs,ygs,x1,y1,G11,G12,G22,T111,T112,T122,T211,T212,T222,gvar)

				//	Loop over all elements of a panel
				for(int m=0; m<rmtElms; m++){

					xgs = xls-0.5*(Xs[m+mp1]+Xs[m]);
					ygs = yls-0.5*(Ys[m+mp1]+Ys[m]);
					gvar = xgs*xgs+ygs*ygs;
					// Loop over Gauss points
					for(int l=0; l<6; l++){
						x1 = 0.5*dx*IGP[l]+xgs;
						y1 = 0.5*dy*IGP[l]+ygs;

						computeGT(1, x1, y1, G11, G12, G22, T111, T112, T122, T211, T212, T222);
						gvar = IGW[l]*r*0.5;

						A[fixind+2*m] -= G11*gvar;
						A[fixind+NOFN+2*m] -= G12*gvar;
						A[fixind+1+2*m] -= G12*gvar;
						A[fixind+NOFN+1+2*m] -= G22*gvar;

						rhs[yMatID+2*fixsize+2*m]   -= dEta*((T111*lnx+T112*lny)*u1+(T112*lnx+T122*lny)*u2)*gvar;
						rhs[yMatID+2*fixsize+2*m+1] -= dEta*((T211*lnx+T212*lny)*u1+(T212*lnx+T222*lny)*u2)*gvar;

					}
				}
			}
# pragma omp barrier
			yMatID += 2*rmtElms;
		}
	}
}



void MatBlC1::SSineFlow(int inP, double qAvel, double wavenumber, double dEta){
	double dx, dy, xls, yls, xgs, ygs, r, lnx, lny, u1, u2;
	double x1, y1, gvar, G11, G12, G22;
	double T111, T122, T112, T211, T222, T212;
	double dxpanel, dypanel, xppanel, yppanel, Vn, dpanel, nxpanel, nypanel, scoord;
	double *Xp, *Yp, *Xs, *Ys;  // Xp for panels and Xs for singularities
	int localElm = bem->allSize[inP];
	int NOFN = nOFm-2*fixsize;
	int rmtElms, fixind, mp1;
	int xMatID = 4*fixsize*fixsize;
	int yMatID = 0;
    
	Xp = bem->allXId[inP];
	Yp = bem->allYId[inP];
    
	for(int m=0; m<inP; m++){
		xMatID += 2*NOFN*bem->allSize[m];
	}
    
	// Determine Panel Parameters
	dypanel = Yp[localElm]-Yp[0];
	dxpanel = Xp[localElm]-Xp[0];
	xppanel = 0.5*(Xp[localElm]+Xp[0]);
	yppanel = 0.5*(Yp[localElm]+Yp[0]);
	dpanel = sqrt(dxpanel*dxpanel+dypanel*dypanel);
	nxpanel = dypanel/dpanel;
    nypanel = -dxpanel/dpanel;

    
	//	Size of the Element, the Position of its center, normals
	for(int n=0; n<localElm; n++){
		dx = Xp[n+1]-Xp[n];
		dy = Yp[n+1]-Yp[n];
		xls = 0.5*(Xp[n+1]+Xp[n]);
		yls = 0.5*(Yp[n+1]+Yp[n]);
		r = (dx*dx+dy*dy);
		r = sqrt(r);
		lnx = dy/r;
		lny = -dx/r;
        
        scoord = -(xls-Xp[0])*nypanel + (yls-Yp[0])*nxpanel;
        Vn = qAvel*cos(2*M_PI*wavenumber*scoord/dpanel);
        
		u1 = Vn*nxpanel;
		u2 = Vn*nypanel;
		yMatID = 0;
        
		//	Loop over all panels, switch for singularity position (midpoint corner)
		for(int k=dynblock; k<nbBlocks; k++){
			rmtElms = bem->Elements[k].getFullSize();
			Xs = bem->Elements[k].XId[0];
			Ys = bem->Elements[k].YId[0];
			fixind = yMatID+xMatID+2*n*NOFN;
			mp1 = (k<bem->ifblock);
#pragma omp parallel
			{
                
#pragma omp for private(xgs,ygs,x1,y1,G11,G12,G22,T111,T112,T122,T211,T212,T222,gvar)
                
				//	Loop over all elements of a panel
				for(int m=0; m<rmtElms; m++){
                    
					xgs = xls-0.5*(Xs[m+mp1]+Xs[m]);
					ygs = yls-0.5*(Ys[m+mp1]+Ys[m]);
					gvar = xgs*xgs+ygs*ygs;
					// Loop over Gauss points
					for(int l=0; l<6; l++){
						x1 = 0.5*dx*IGP[l]+xgs;
						y1 = 0.5*dy*IGP[l]+ygs;
                        
						computeGT(1, x1, y1, G11, G12, G22, T111, T112, T122, T211, T212, T222);
						gvar = IGW[l]*r*0.5;
                        
						A[fixind+2*m] -= G11*gvar;
						A[fixind+NOFN+2*m] -= G12*gvar;
						A[fixind+1+2*m] -= G12*gvar;
						A[fixind+NOFN+1+2*m] -= G22*gvar;
                        
						rhs[yMatID+2*fixsize+2*m]   -= dEta*((T111*lnx+T112*lny)*u1+(T112*lnx+T122*lny)*u2)*gvar;
						rhs[yMatID+2*fixsize+2*m+1] -= dEta*((T211*lnx+T212*lny)*u1+(T212*lnx+T222*lny)*u2)*gvar;
                        
					}
				}
			}
# pragma omp barrier
			yMatID += 2*rmtElms;
		}
	}
}

void MatBlC1::SCircTrac(int inP, double delT){
	double dx, dy, xls, yls, xgs, ygs, r, lnx, lny;
	double x1, y1, G11, G12, G22, fl1, fl2, fr1, fr2;
	double T111, T122, T112, T211, T222, T212,f1,f2;
	double phil, phir, tx1, ty1, tx3, ty3, s1, s3, d1, d2, unorm, unorp;
	double dLambda = Lambda - 1.0;
	double *Xp, *Yp, *Xs, *Ys;  // Xp for panels and Xs for singularities
	int localElm = bem->allSize[inP];
	int rmtElms, nm1, np1, np2, dolog, mp1;
	int NOFN = nOFm-2*fixsize;
	int MOFM = 2*fixsize;
	//    double fifac = 2*kF*pow(1-Eta,0.66666)*pow(nGamma,0.33333)/(3*sqrt(12));
	int xMatID1 = 0;
	int ind1 = 2*fixsize*(2*fixsize+NOFN);
	int xMatID2 = 0;
	int ind2 = 2*fixsize*(2*fixsize+2*NOFN);
	int yMatID = 0;
	double LMgam, LPgam, RMgam,RPgam,kappa;
	double maxForce = 0;
	double minForce = 1e9;

	double alpha = 1.0;
	if(kF < 1e-3){
		alpha=0.0;
	}

	Xp = bem->allXId[inP];
	Yp = bem->allYId[inP];

	int fixind;

	for(int m=dynpanel; m<inP; m++){
		xMatID1 += 2*fixsize*bem->allSize[m];
		xMatID2 += 2*bem->allSize[m];
	}

	//	Size of the Element, the Position of its center, normals
	for(int n=0; n<localElm; n++){
		nm1 = n-1;
		np1 = n+1;
		np2 = n+2;
		if (n==localElm-1) {
			np1 = 0;
			np2 = 1;
		} else if(n==0){
			nm1 = localElm-1;
		}else if(n==localElm-2){
			np2=0;
		}

		tx1 = Xp[n]-Xp[nm1];
		ty1 = Yp[n]-Yp[nm1];
		s1 = sqrt(tx1*tx1+ty1*ty1);
		tx3 = Xp[np2]-Xp[np1];
		ty3 = Yp[np2]-Yp[np1];
		s3 = sqrt(tx3*tx3+ty3*ty3);
		dx = Xp[np1]-Xp[n];
		dy = Yp[np1]-Yp[n];
		r = sqrt(dx*dx+dy*dy);
		d1 = s1+r;
		d2 = s3+r;

		xls = 0.5*(Xp[n+1]+Xp[n]);
		yls = 0.5*(Yp[n+1]+Yp[n]);
		lnx = dy/r;
		lny = -dx/r;

		LMgam = LocalGam(inP, nm1, n)-LocalGam(inP, n, n)*(1.0-M_PI_4)*alpha;
		LPgam = LocalGam(inP, n, np1)-LocalGam(inP, n, n)*(1.0-M_PI_4)*alpha;
		RMgam = LocalGam(inP, n, np1)-LocalGam(inP, np1, np1)*(1.0-M_PI_4)*alpha;
		RPgam = LocalGam(inP, np1, np2)-LocalGam(inP, np1, np1)*(1.0-M_PI_4)*alpha;

		fl1 = 2.0*(LPgam*dx/r-LMgam*tx1/s1)/d1;
		fl2 = 2.0*(LPgam*dy/r-LMgam*ty1/s1)/d1;
		fr1 = 2.0*(RPgam*tx3/s3-RMgam*dx/r)/d2;
		fr2 = 2.0*(RPgam*ty3/s3-RMgam*dy/r)/d2;

		unorm = getUvel(inP, nm1, n, np1);
		kappa = LocalGam(inP, n, n);
		(this->*applybc)(Xp[n], Yp[n], lnx, lny, unorm, kappa, f1, f2);
		fl1 +=  f1;
		fl2 +=  f2;

		unorp = getUvel(inP, n, np1, np2);
		kappa = LocalGam(inP, np1, np1);
		(this->*applybc)(Xp[n+1], Yp[n+1], lnx, lny, unorp, kappa, f1, f2);
		fr1 +=  f1;
		fr2 +=  f2;

		fl1 += SpecialForces(inP,n)*lnx;
		fl2 += SpecialForces(inP,n)*lny;
		fr1 += SpecialForces(inP,np1)*lnx;
		fr2 += SpecialForces(inP,np1)*lny;

		maxForce = std::max(sqrt(fl1*fl1+fl2*fl2),maxForce);
		minForce = std::min(sqrt(fl1*fl1+fl2*fl2),minForce);

		yMatID = 0;

		A[ind2+(2*n+xMatID2)*(NOFN+1)] -= 2.0*M_PI*(Lambda + 1.0);
		A[ind2+(2*n+1+xMatID2)*(NOFN+1)] -= 2.0*M_PI*(Lambda + 1.0);

		rhs[2*fixsize+xMatID2+2*n]   += (fl1+fr1)*0.25*r*(2*log(r)-1)-0.5*r*fl1;
		rhs[2*fixsize+ xMatID2+2*np1]   += (fl1+fr1)*0.25*r*(2*log(r)-1)-0.5*r*fr1;
		// Including fr1 might cause problems in BordTrac
		rhs[2*fixsize+ xMatID2+2*n+1] += (fl2+fr2)*0.25*r*(2*log(r)-1)-0.5*r*fl2;
		rhs[2*fixsize+ xMatID2+2*np1+1] += (fl2+fr2)*0.25*r*(2*log(r)-1)-0.5*r*fr2;
		// Integration is alright the 3/2 term is obtained due to the last term

		phil = -delT*r*0.25*(2*log(r)-3);
		phir = -delT*r*0.25*(2*log(r)-1);

		A[ind2+xMatID2+2*n+NOFN*(2*nm1+xMatID2)] += LMgam*phil/(s1*d1);
		A[ind2+xMatID2+2*np1+NOFN*(2*nm1+xMatID2)] += LMgam*phir/(s1*d1);

		A[ind2+(2*n+xMatID2)*(NOFN+1)] -= phil*(LMgam/(s1*d1)+LPgam/(r*d1))-RMgam*phir/(r*d2);
		A[ind2+2*np1+xMatID2+NOFN*(xMatID2+2*n)] -= phir*(LMgam/(s1*d1)+LPgam/(r*d1))-RMgam*phil/(r*d2);

		A[ind2+2*n+xMatID2+NOFN*(xMatID2+2*np1)] -= phir*(RMgam/(r*d2)+RPgam/(s3*d2))-LPgam*phil/(r*d1);
		A[ind2+2*np1+xMatID2+NOFN*(xMatID2+2*np1)] -= phil*(RMgam/(r*d2)+RPgam/(s3*d2))-LPgam*phir/(r*d1);

		A[ind2+2*n+xMatID2+NOFN*(2*np2+xMatID2)] += RPgam*phir/(s3*d2);
		A[ind2+2*np1+xMatID2+NOFN*(2*np2+xMatID2)] += RPgam*phil/(s3*d2);

		A[ind2+xMatID2+2*n+1+NOFN*(xMatID2+2*nm1+1)] += LMgam*phil/(s1*d1);
		A[ind2+xMatID2+2*np1+1+NOFN*(xMatID2+2*nm1+1)] += LMgam*phir/(s1*d1);

		A[ind2+(xMatID2+2*n+1)*(NOFN+1)] -= phil*(LMgam/(s1*d1)+LPgam/(r*d1))-RMgam*phir/(r*d2);
		A[ind2+xMatID2+2*np1+1+NOFN*(xMatID2+2*n+1)] -= phir*(LMgam/(s1*d1)+LPgam/(r*d1))-RMgam*phil/(r*d2);

		A[ind2+xMatID2+2*n+1+NOFN*(xMatID2+2*np1+1)] -= phir*(RMgam/(r*d2)+RPgam/(s3*d2))-LPgam*phil/(r*d1);
		A[ind2+xMatID2+2*np1+1+NOFN*(xMatID2+2*np1+1)] -= phil*(RMgam/(r*d2)+RPgam/(s3*d2))-LPgam*phir/(r*d1);

		A[ind2+xMatID2+2*n+1+NOFN*(2*np2+xMatID2+1)] += RPgam*phir/(s3*d2);
		A[ind2+xMatID2+2*np1+1+NOFN*(2*np2+xMatID2+1)] += RPgam*phil/(s3*d2);

		//	Loop over all panels, switch for singularity position (midpoint corner)
		for(int k=0; k<dynblock; k++){
			rmtElms = bem->Elements[k].getFullSize();
			Xs = bem->Elements[k].XId[0];
			Ys = bem->Elements[k].YId[0];
			fixind = ind1 + yMatID+xMatID1;

#pragma omp parallel
			{
#pragma omp for private(xgs,ygs,x1,y1,G11,G12,G22,T111,T112,T122,T211,T212,T222,phil,phir)
				//	Loop over all elements of a panel
				for(int m=0; m<rmtElms; m++){

					xgs = xls-0.5*(Xs[m+1]+Xs[m]);
					ygs = yls-0.5*(Ys[m+1]+Ys[m]);

					// Loop over Gauss point

					for(int l=0; l<6; l++){
						x1 = 0.5*dx*IGP[l]+xgs;
						y1 = 0.5*dy*IGP[l]+ygs;

						computeGT(1, x1, y1, G11, G12, G22, T111, T112, T122, T211, T212, T222);

						phil = (1.0-IGP[l])*0.25*IGW[l]*r;
						phir = (1.0+IGP[l])*0.25*IGW[l]*r;

						rhs[yMatID+2*m]   += (fl1*phil+fr1*phir)*G11+(fl2*phil+fr2*phir)*G12;
						rhs[yMatID+2*m+1] += (fl1*phil+fr1*phir)*G12+(fl2*phil+fr2*phir)*G22;

						A[fixind+2*n*MOFM+2*m]  += dLambda*phil*(T111*lnx+T112*lny);    // eq:m1 co:ul
						A[fixind+(2*n+1)*MOFM+2*m] += dLambda*phil*(T112*lnx+T122*lny); // eq:m1 co:vl
						A[fixind+MOFM*2*np1+2*m]	   += dLambda*phir*(T111*lnx+T112*lny); // eq:m1 co:ur
						A[fixind+MOFM*(2*np1+1)+2*m]  += dLambda*phir*(T112*lnx+T122*lny); // eq:m1 co:vr

						A[fixind+2*n*MOFM+1+2*m] += dLambda*phil*(T211*lnx+T212*lny); // eq:m2 co:ul
						A[fixind+2*n*MOFM+1+MOFM+2*m] += dLambda*phil*(T212*lnx+T222*lny); // eq:m2 co:vl
						A[fixind+1+MOFM*(2*np1)+2*m]  += dLambda*phir*(T211*lnx+T212*lny); // eq:m2 co:ur
						A[fixind+1+MOFM*(2*np1+1)+2*m]+= dLambda*phir*(T212*lnx+T222*lny); // eq:m2 c0:vr

						// Implicit Interface

						phil = delT*phil;
						phir = delT*phir;

						A[fixind+2*nm1*MOFM+2*m]	  += -G11*phil/(s1*d1)*LMgam; // eq:m1 co:ull
						A[fixind+(2*nm1+1)*MOFM+2*m] += -G12*phil/(s1*d1)*LMgam; // eq:m1 co:vll

						A[fixind+2*n*MOFM+2*m]	  += G11*(phil*(LMgam/(s1*d1)+LPgam/(r*d1))-RMgam*phir/(r*d2)); // eq:m1 co:ul
						A[fixind+(2*n+1)*MOFM+2*m]	  += G12*(phil*(LMgam/(s1*d1)+LPgam/(r*d1))-RMgam*phir/(r*d2)); // eq:m1 co:vl

						A[fixind+2*np1*MOFM+2*m]	  += G11*(phir*(RMgam/(r*d2)+RPgam/(s3*d2))-LPgam*phil/(r*d1));
						A[fixind+(2*np1+1)*MOFM+2*m] += G12*(phir*(RMgam/(r*d2)+RPgam/(s3*d2))-LPgam*phil/(r*d1));

						A[fixind+2*np2*MOFM+2*m]	  += -G11*phir/(s3*d2)*RPgam;
						A[fixind+(2*np2+1)*MOFM+2*m] += -G12*phir/(s3*d2)*RPgam;


						A[fixind+1+2*nm1*MOFM+2*m]   += -G12*phil/(s1*d1)*LMgam; // eq:m2 co:ull
						A[fixind+1+(2*nm1+1)*MOFM+2*m] += -G22*phil/(s1*d1)*LMgam; // eq:m2 co:vll

						A[fixind+1+2*n*MOFM+2*m]		+= G12*(phil*(LMgam/(s1*d1)+LPgam/(r*d1))-RMgam*phir/(r*d2));
						A[fixind+1+(2*n+1)*MOFM+2*m]	+= G22*(phil*(LMgam/(s1*d1)+LPgam/(r*d1))-RMgam*phir/(r*d2));

						A[fixind+1+2*np1*MOFM+2*m]   += G12*(phir*(RMgam/(r*d2)+RPgam/(s3*d2))-LPgam*phil/(r*d1));
						A[fixind+1+(2*np1+1)*MOFM+2*m] += G22*(phir*(RMgam/(r*d2)+RPgam/(s3*d2))-LPgam*phil/(r*d1));

						A[fixind+1+2*np2*MOFM+2*m]	+= -G12*phir/(s3*d2)*RPgam;
						A[fixind+1+(2*np2+1)*MOFM+2*m]	+= -G22*phir/(s3*d2)*RPgam;

					}
				}
			}
# pragma omp barrier
			yMatID += 2*rmtElms;
		}
		yMatID =0;
		for(int k=dynblock; k<nbBlocks; k++){
			rmtElms = bem->Elements[k].getFullSize();
			Xs = bem->Elements[k].XId[0];
			Ys = bem->Elements[k].YId[0];
			fixind = ind2+yMatID+ xMatID2*NOFN;
			mp1 = (k<bem->ifblock);
#pragma omp parallel
			{
#pragma omp for private(xgs,ygs,x1,y1,G11,G12,G22,T111,T112,T122,T211,T212,T222,phil,phir,dolog)
				//	Loop over all elements of a panel
				for(int m=0; m<rmtElms; m++){
					xgs = xls-0.5*(Xs[m+mp1]+Xs[m]);
					ygs = yls-0.5*(Ys[m+mp1]+Ys[m]);

					// Decide wheather it is a singular element
					dolog = (!(((n==m)||(np1==m))&&(k==(inP-dynpanel+dynblock))));

					// Loop over Gauss points
					for(int l=0; l<6; l++){
						x1 = 0.5*dx*IGP[l]+xgs;
						y1 = 0.5*dy*IGP[l]+ygs;

						computeGT(dolog, x1, y1, G11, G12, G22, T111, T112, T122, T211, T212, T222);

						phil = (1.0-IGP[l])*0.25*IGW[l]*r;
						phir = (1.0+IGP[l])*0.25*IGW[l]*r;

						rhs[yMatID+2*fixsize+2*m]   += (fl1*phil+fr1*phir)*G11+(fl2*phil+fr2*phir)*G12;
						rhs[yMatID+2*fixsize+2*m+1] += (fl1*phil+fr1*phir)*G12+(fl2*phil+fr2*phir)*G22;

						A[fixind+2*n*NOFN+2*m]  += dLambda*phil*(T111*lnx+T112*lny);
						A[fixind+(2*n+1)*NOFN+2*m] += dLambda*phil*(T112*lnx+T122*lny);
						A[fixind+NOFN*2*np1+2*m]	   += dLambda*phir*(T111*lnx+T112*lny);
						A[fixind+NOFN*(2*np1+1)+2*m]  += dLambda*phir*(T112*lnx+T122*lny);

						A[fixind+1+2*n*NOFN+2*m] += dLambda*phil*(T211*lnx+T212*lny);
						A[fixind+1+(2*n+1)*NOFN+2*m] += dLambda*phil*(T212*lnx+T222*lny);
						A[fixind+1+NOFN*(2*np1)+2*m]  += dLambda*phir*(T211*lnx+T212*lny);
						A[fixind+1+NOFN*(2*np1+1)+2*m]+= dLambda*phir*(T212*lnx+T222*lny);


						// Implicit Interface

						phil = delT*phil;
						phir = delT*phir;

						A[fixind+2*nm1*NOFN+2*m]	  += -G11*phil/(s1*d1)*LMgam;
						A[fixind+(2*nm1+1)*NOFN+2*m] += -G12*phil/(s1*d1)*LMgam;

						A[fixind+2*n*NOFN+2*m]	  += G11*(phil*(LMgam/(s1*d1)+LPgam/(r*d1))-RMgam*phir/(r*d2));
						A[fixind+(2*n+1)*NOFN+2*m]	  += G12*(phil*(LMgam/(s1*d1)+LPgam/(r*d1))-RMgam*phir/(r*d2));

						A[fixind+2*np1*NOFN+2*m]	  += G11*(phir*(RMgam/(r*d2)+RPgam/(s3*d2))-LPgam*phil/(r*d1));
						A[fixind+(2*np1+1)*NOFN+2*m] += G12*(phir*(RMgam/(r*d2)+RPgam/(s3*d2))-LPgam*phil/(r*d1));

						A[fixind+2*np2*NOFN+2*m]	  += -G11*phir/(s3*d2)*RPgam;
						A[fixind+(2*np2+1)*NOFN+2*m] += -G12*phir/(s3*d2)*RPgam;


						A[fixind+1+2*nm1*NOFN+2*m]   += -G12*phil/(s1*d1)*LMgam;
						A[fixind+1+(2*nm1+1)*NOFN+2*m] += -G22*phil/(s1*d1)*LMgam;

						A[fixind+1+2*n*NOFN+2*m]		+= G12*(phil*(LMgam/(s1*d1)+LPgam/(r*d1))-RMgam*phir/(r*d2));
						A[fixind+1+(2*n+1)*NOFN+2*m]	+= G22*(phil*(LMgam/(s1*d1)+LPgam/(r*d1))-RMgam*phir/(r*d2));

						A[fixind+1+2*np1*NOFN+2*m]   += G12*(phir*(RMgam/(r*d2)+RPgam/(s3*d2))-LPgam*phil/(r*d1));
						A[fixind+1+(2*np1+1)*NOFN+2*m] += G22*(phir*(RMgam/(r*d2)+RPgam/(s3*d2))-LPgam*phil/(r*d1));

						A[fixind+1+2*np2*NOFN+2*m]	+= -G12*phir/(s3*d2)*RPgam;
						A[fixind+1+(2*np2+1)*NOFN+2*m]	+= -G22*phir/(s3*d2)*RPgam;

					}
				}
			}
# pragma omp barrier
			yMatID += 2*rmtElms;
		}
	}

	dominantForce = std::max(maxForce-minForce,dominantForce);
}

// void MatBlC1::SBordTrac(int inP, double delT){

// ------------------- Dedicated to Visula reconstruction, their routines should match the corresponding problem routines


void MatBlC1::VisuZeroVel(int inP, double *Xpos, double *Ypos, int points, double *U, double *V, double *P){
    double dx, dy, xls, yls, xgs, ygs, r, f1, f2;
	double x1, y1, gvar, G11, G12, G22, M1, M2, P11, P12, P22;
    double *Xp, *Yp;
    int localElm = bem->allSize[inP];
	int xMatID = 0;
    
	Xp = bem->allXId[inP];
    Yp = bem->allYId[inP];
    
	for(int m=0; m<inP; m++){
		xMatID += 2*bem->allSize[m];
	}
	
	//	Size of the Element, the Position of its center, Normals
	for(int n=0; n<localElm; n++){
		dx = Xp[n+1]-Xp[n];
        dy = Yp[n+1]-Yp[n];
        xls = 0.5*(Xp[n+1]+Xp[n]);
        yls = 0.5*(Yp[n+1]+Yp[n]);
        r = (dx*dx+dy*dy);
        r = sqrt(r);
        f1 = rhs[xMatID+2*n];
        f2 = rhs[xMatID+2*n+1];
        
#pragma omp parallel
        {
#pragma omp for private(xgs,ygs,x1,y1,G11,G12,G22, M1, M2,gvar)
            //	Loop over all panels, switch for singularity position (midpoint corner)
            for(int k=0; k<points; k++){
                
				//	Loop over all elements of a panel
                xgs = xls-Xpos[k];
				ygs = yls-Ypos[k];
                
                // Loop over Gauss points
                for(int l=0; l<6; l++){
                    x1 = 0.5*dx*IGP[l]+xgs;
                    y1 = 0.5*dy*IGP[l]+ygs;
                    
                    computeGO(1, x1, y1, G11, G12, G22);
                    computeMP(x1, y1, M1, M2, P11, P12, P22);
                    gvar = IGW[l]*r*0.5;
                    
                    U[k] -= (G11*f1 + G12*f2)*gvar;
                    V[k] -= (G12*f1 + G22*f2)*gvar;
                    P[k] -= (f1*M1+f2*M2)*gvar;
                }
				
			}
		}
# pragma omp barrier
    }
}

void MatBlC1::VisuFlatVel(int inP, double *Xpos, double *Ypos, double Vn, double Vt, double Theta, double dEta, int points, double *U, double *V, double *P){
	
	/*	Integration from all singularities over a boundary with n-t velocity.
	 Wall Elements are always zeroth order
	 */
	
	double dx, dy, xls, yls, xgs, ygs, gvar, r, lnx, lny, u1, u2, xcm, ycm;
	double x1, y1, G11, G12, G22, T111, T122, T112, T211, T222, T212, M1, M2, P11, P12, P22, f1, f2;
	
	double *Xp, *Yp;
	int localElm = bem->allSize[inP];
	int xMatID = 0;
	
	Xp = bem->allXId[inP];
    Yp = bem->allYId[inP];
    
	for(int m=0; m<inP; m++){
		xMatID += 2*bem->allSize[m];
	}
	
    bem->getCoM(bem->panel2block[inP], xcm, ycm);
    
	//	Size of the Element, the Position of its center, normals
	for(int n=0; n<localElm; n++){
		dx = Xp[n+1]-Xp[n];
        dy = Yp[n+1]-Yp[n];
        xls = 0.5*(Xp[n+1]+Xp[n]);
        yls = 0.5*(Yp[n+1]+Yp[n]);
        r = (dx*dx+dy*dy);
        r = sqrt(r);
        lnx = dy/r;
        lny = -dx/r;
        u1 = Vn - (yls-ycm)*Theta;
        u2 = Vt + (xls-xcm)*Theta;
        f1 = rhs[xMatID+2*n];
        f2 = rhs[xMatID+2*n+1];
        
#pragma omp parallel
        {
#pragma omp for private(xgs,ygs,x1,y1,G11,G12,G22,T111,T112,T122,T211,T212,T222,gvar, M1, M2, P11, P12, P22)
            //	Loop over all panels, switch for singularity position (midpoint corner)
            for(int k=0; k<points; k++){
                //	Loop over all elements of a panel
				xgs = xls-Xpos[k];
                ygs = yls-Ypos[k];
                
                // Loop over Gauss points
                for(int l=0; l<6; l++){
					x1 = 0.5*dx*IGP[l]+xgs;
                    y1 = 0.5*dy*IGP[l]+ygs;
                    
                    computeGT(1, x1, y1, G11, G12, G22, T111, T112, T122, T211, T212, T222);
                    computeMP(x1, y1, M1, M2, P11, P12, P22);
                    gvar = IGW[l]*r*0.5;
                    
                    U[k] += (-G11*f1 -G12*f2 +dEta*((T111*lnx+T112*lny)*u1+(T112*lnx+T122*lny)*u2))*gvar;
                    V[k] += (-G12*f1 -G22*f2 +dEta*((T211*lnx+T212*lny)*u1+(T212*lnx+T222*lny)*u2))*gvar;
                    P[k] += (-M1*f1 -M2*f2 +dEta*((P11*lnx+P12*lny)*u1+(P12*lnx+P22*lny)*u2))*gvar;
				}
			}
		}
# pragma omp barrier
	}
}

void MatBlC1::VisuNormVel(int inP, double *Xpos, double *Ypos, double Vn, double Vt, double dEta, int points, double *U, double *V, double *P){
	
	/*	Integration from all singularities over a boundary with n-t velocity.
	 Wall Elements are always zeroth order
	 */
	
	double dx, dy, xls, yls, xgs, ygs, gvar, r, lnx, lny, u1, u2;
	double x1, y1, G11, G12, G22, T111, T122, T112, T211, T222, T212, M1, M2, P11, P12, P22, f1, f2;
	
	double *Xp, *Yp;
	int localElm = bem->allSize[inP];
	int xMatID = 0;
	
	Xp = bem->allXId[inP];
    Yp = bem->allYId[inP];
    
	for(int m=0; m<inP; m++){
		xMatID += 2*bem->allSize[m];
	}
	
	//	Size of the Element, the Position of its center, normals
	for(int n=0; n<localElm; n++){
		dx = Xp[n+1]-Xp[n];
        dy = Yp[n+1]-Yp[n];
        xls = 0.5*(Xp[n+1]+Xp[n]);
        yls = 0.5*(Yp[n+1]+Yp[n]);
        r = (dx*dx+dy*dy);
        r = sqrt(r);
        lnx = dy/r;
        lny = -dx/r;
        u1 = Vn*lnx-Vt*lny;
        u2 = Vn*lny+Vt*lnx;
        f1 = rhs[xMatID+2*n];
        f2 = rhs[xMatID+2*n+1];
        
#pragma omp parallel
        {
#pragma omp for private(xgs,ygs,x1,y1,G11,G12,G22,T111,T112,T122,T211,T212,T222,gvar, M1, M2, P11, P12, P22)
            //	Loop over all panels, switch for singularity position (midpoint corner)
            for(int k=0; k<points; k++){
                
                
				//	Loop over all elements of a panel
				xgs = xls-Xpos[k];
                ygs = yls-Ypos[k];
                
                // Loop over Gauss points
                for(int l=0; l<6; l++){
					x1 = 0.5*dx*IGP[l]+xgs;
                    y1 = 0.5*dy*IGP[l]+ygs;
                    
                    computeGT(1, x1, y1, G11, G12, G22, T111, T112, T122, T211, T212, T222);
                    computeMP(x1, y1, M1, M2, P11, P12, P22);
                    gvar = IGW[l]*r*0.5;
                    
                    U[k] += ( -G11*f1 -G12*f2 +dEta*((T111*lnx+T112*lny)*u1+(T112*lnx+T122*lny)*u2))*gvar;
                    V[k] += ( -G12*f1 -G22*f2 +dEta*((T211*lnx+T212*lny)*u1+(T212*lnx+T222*lny)*u2))*gvar;
                    P[k] += ( -M1*f1 -M2*f2 +dEta*((P11*lnx+P12*lny)*u1+(P12*lnx+P22*lny)*u2))*gvar;
				}
			}
		}
# pragma omp barrier
	}
}

void MatBlC1::VisuFiniFlow(int inP, double *Xpos, double *Ypos, double Vn, double Vt, double dEta, int points, double *U, double *V, double *P){
	
	/*	Integration from all singularities over a boundary with n-t velocity.
	 Wall Elements are always zeroth order
	 */
	
	double dx, dy, xls, yls, xgs, ygs, gvar, r, lnx, lny, u1, u2;
	double x1, y1, G11, G12, G22, T111, T122, T112, T211, T222, T212, M1, M2, P11, P12, P22, f1, f2;
	
	double *Xp, *Yp;
	int localElm = bem->allSize[inP];
	int xMatID = 0;
	
	Xp = bem->allXId[inP];
    Yp = bem->allYId[inP];
    
	for(int m=0; m<inP; m++){
		xMatID += 2*bem->allSize[m];
	}
	
	//	Size of the Element, the Position of its center, normals
	for(int n=0; n<localElm; n++){
		dx = Xp[n+1]-Xp[n];
        dy = Yp[n+1]-Yp[n];
        xls = 0.5*(Xp[n+1]+Xp[n]);
        yls = 0.5*(Yp[n+1]+Yp[n]);
        r = (dx*dx+dy*dy);
        r = sqrt(r);
        lnx = dy/r;
        lny = -dx/r;
        u1 = xls;
        u2 = -yls;
        f1 = rhs[xMatID+2*n];
        f2 = rhs[xMatID+2*n+1];
        
#pragma omp parallel
        {
#pragma omp for private(xgs,ygs,x1,y1,G11,G12,G22,T111,T112,T122,T211,T212,T222,gvar, M1, M2, P11, P12, P22)
            //	Loop over all panels, switch for singularity position (midpoint corner)
            for(int k=0; k<points; k++){
                
                
				//	Loop over all elements of a panel
				xgs = xls-Xpos[k];
                ygs = yls-Ypos[k];
                
                // Loop over Gauss points
                for(int l=0; l<6; l++){
					x1 = 0.5*dx*IGP[l]+xgs;
                    y1 = 0.5*dy*IGP[l]+ygs;
                    
                    computeGT(1, x1, y1, G11, G12, G22, T111, T112, T122, T211, T212, T222);
                    computeMP(x1, y1, M1, M2, P11, P12, P22);
                    gvar = IGW[l]*r*0.5;
                    
                    U[k] += (-G11*f1 - G12*f2 +dEta*((T111*lnx+T112*lny)*u1+(T112*lnx+T122*lny)*u2))*gvar;
                    V[k] += (-G12*f1 - G22*f2 +dEta*((T211*lnx+T212*lny)*u1+(T212*lnx+T222*lny)*u2))*gvar;
                    P[k] += (-M1*f1 - M2*f2 +dEta*((P11*lnx+P12*lny)*u1+(P12*lnx+P22*lny)*u2))*gvar;
				}
			}
		}
# pragma omp barrier
	}
}

void MatBlC1::VisuPressure(int inP, double *Xpos, double *Ypos, double plevel, double dEta, int points, double *U, double *V, double *P){
    double dx, dy, xls, yls, xgs, ygs, r, lnx, lny, f1, f2, u1 ,u2;
	double x1, y1, G11, G12, G22, gvar;
	double T111, T122, T112, T211, T222, T212, M1, M2, P11, P12, P22;
    double *Xp, *Yp;
	int localElm = bem->allSize[inP];
    
	int xMatID = 0;
    
	
	Xp = bem->allXId[inP];
    Yp = bem->allYId[inP];
	
	for(int m=0; m<inP; m++){
		xMatID += 2*bem->allSize[m];
	}
	
	//	Size of the Element, the Position of its center, Normals
	for(int n=0; n<localElm; n++){
		dx = Xp[n+1]-Xp[n];
        dy = Yp[n+1]-Yp[n];
        xls = 0.5*(Xp[n+1]+Xp[n]);
        yls = 0.5*(Yp[n+1]+Yp[n]);
        r = dx*dx+dy*dy;
        r = sqrt(r);
        lnx = dy/r;
        lny = -dx/r;
        f1 = -plevel*lnx;
        f2 = -plevel*lny;
        u1 = rhs[xMatID+2*n];
        u2 = rhs[xMatID+2*n+1];
        
#pragma omp parallel
        {
#pragma omp for private(xgs,ygs,x1,y1,G11,G12,G22,T111,T112,T122,T211,T212,T222,gvar,M1, M2,P11,P12,P22)
            //	Loop over all panels, switch for singularity position (midpoint corner)
            for(int k=0; k<points; k++){
                xgs = xls-Xpos[k];
                ygs = yls-Ypos[k];
                
                
                
                // Loop over Gauss points
				for(int l=0; l<6; l++){
					x1 = 0.5*dx*IGP[l]+xgs;
					y1 = 0.5*dy*IGP[l]+ygs;
                    
					computeGT(1, x1, y1, G11, G12, G22, T111, T112, T122, T211, T212, T222);
                    computeMP(x1, y1, M1, M2, P11, P12, P22);
					gvar = IGW[l]*r*0.5;
                    
                    U[k] += ( dEta*((T111*lnx+T112*lny)*u1 + (T112*lnx+T122*lny)*u2) -f1*G11-f2*G12)*gvar;
                    V[k] += ( dEta*((T211*lnx+T212*lny)*u1 + (T212*lnx+T222*lny)*u2) -f1*G12-f2*G22)*gvar;
                    P[k] += ( dEta*((P11*lnx+P12*lny)*u1 + (P12*lnx+P22*lny)*u2) -f1*M1-f2*M2)*gvar;
                    
                    
                }
            }
		}
# pragma omp barrier
	}
}

void MatBlC1::VisuPOX(int inP, double *Xpos, double *Ypos, double plevel, double dEta, int points, double *U, double *V, double *P){
    // Works only if wall is aligned with y axis
    
    double dx, dy, xls, yls, xgs, ygs, r, lnx, lny, f1, f2, u1 ,u2;
	double x1, y1, G11, G12, G22, gvar;
	double T111, T122, T112, T211, T222, T212, M1, M2, P11, P12, P22;
    double *Xp, *Yp;
	int localElm = bem->allSize[inP];
    
	int xMatID = 0;
    
	
	Xp = bem->allXId[inP];
    Yp = bem->allYId[inP];
	
	for(int m=0; m<inP; m++){
		xMatID += 2*bem->allSize[m];
	}
	
	//	Size of the Element, the Position of its center, Normals
	for(int n=0; n<localElm; n++){
		dx = Xp[n+1]-Xp[n];
        dy = Yp[n+1]-Yp[n];
        xls = 0.5*(Xp[n+1]+Xp[n]);
        yls = 0.5*(Yp[n+1]+Yp[n]);
        r = dx*dx+dy*dy;
        r = sqrt(r);
        lnx = dy/r;
        lny = -dx/r;
        f1 = plevel;
        f2 = rhs[xMatID+2*n+1];
        u1 = rhs[xMatID+2*n];
        u2 = 0.0;
        
#pragma omp parallel
        {
#pragma omp for private(xgs,ygs,x1,y1,G11,G12,G22,T111,T112,T122,T211,T212,T222,gvar,M1, M2,P11,P12,P22)
            //	Loop over all panels, switch for singularity position (midpoint corner)
            for(int k=0; k<points; k++){
                xgs = xls-Xpos[k];
                ygs = yls-Ypos[k];
                
                // Loop over Gauss points
				for(int l=0; l<6; l++){
					x1 = 0.5*dx*IGP[l]+xgs;
					y1 = 0.5*dy*IGP[l]+ygs;
                    
					computeGT(1, x1, y1, G11, G12, G22, T111, T112, T122, T211, T212, T222);
                    computeMP(x1, y1, M1, M2, P11, P12, P22);
					gvar = IGW[l]*r*0.5;
                    
                    U[k] += ( dEta*((T111*lnx+T112*lny)*u1 + (T112*lnx+T122*lny)*u2) -f1*G11-f2*G12)*gvar;
                    V[k] += ( dEta*((T211*lnx+T212*lny)*u1 + (T212*lnx+T222*lny)*u2) -f1*G12-f2*G22)*gvar;
                    P[k] += ( dEta*((P11*lnx+P12*lny)*u1 + (P12*lnx+P22*lny)*u2) -f1*M1-f2*M2)*gvar;
                    
                    
                }
            }
		}
# pragma omp barrier
	}
}


void MatBlC1::VisuPoiseuille(int inP, double *Xpos, double *Ypos, double qdebit, double dEta, int points, double *U, double *V, double *P){
	double dx, dy, xls, yls, xgs, ygs, r, lnx, lny, u1, u2, f1, f2;
	double x1, y1, gvar, G11, G12, G22, M1, M2;
	double T111, T122, T112, T211, T222, T212, P11, P12, P22;
	double dxpanel, dypanel, xppanel, yppanel, vel, Vn, dpanel, coshb, pmr;
	double *Xp, *Yp;
	int localElm = bem->allSize[inP];
    
	int xMatID = 0;
    
	
	Xp = bem->allXId[inP];
    Yp = bem->allYId[inP];
    
	for(int m=0; m<inP; m++){
		xMatID += 2*bem->allSize[m];
	}
	
	// Determine Panel Parameters
	dypanel = Yp[localElm]-Yp[0];
	dxpanel = Xp[localElm]-Xp[0];
	xppanel = 0.5*(Xp[localElm]+Xp[0]);
	yppanel = 0.5*(Yp[localElm]+Yp[0]);
	dpanel = sqrt(dxpanel*dxpanel+dypanel*dypanel);
	coshb = cosh(kF*dpanel*0.5);
	vel = qdebit*dpanel*kF/(dpanel*kF*coshb-2*sinh(kF*dpanel/2));
	
	//	Size of the Element, the Position of its center, normals
	for(int n=0; n<localElm; n++){
		dx = Xp[n+1]-Xp[n];
        dy = Yp[n+1]-Yp[n];
        xls = 0.5*(Xp[n+1]+Xp[n]);
        yls = 0.5*(Yp[n+1]+Yp[n]);
        r = (dx*dx+dy*dy);
        r = sqrt(r);
        lnx = dy/r;
        lny = -dx/r;
		pmr = sqrt((xls-xppanel)*(xls-xppanel)+(yls-yppanel)*(yls-yppanel));
		if(kF!=0){
			Vn = vel*(coshb-cosh(kF*pmr));
		} else {
			Vn = qdebit/dpanel*1.5*(1.0-4*(pmr/dpanel)*(pmr/dpanel));
		}
		u1 = -Vn*lnx;
        u2 = -Vn*lny;
        f1 = rhs[xMatID+2*n];
        f2 = rhs[xMatID+2*n+1];
        
#pragma omp parallel
        {
#pragma omp for private(xgs,ygs,x1,y1,G11,G12,G22,T111,T112,T122,T211,T212,T222,gvar,M1, M2, P11,P12,P22 )
            //	Loop over all panels, switch for singularity position (midpoint corner)
            for(int k=0; k<points; k++){
                xgs = xls-Xpos[k];
                ygs = yls-Ypos[k];
				// Loop over Gauss points
				for(int l=0; l<6; l++){
					x1 = 0.5*dx*IGP[l]+xgs;
					y1 = 0.5*dy*IGP[l]+ygs;
                    
					computeGT(1, x1, y1, G11, G12, G22, T111, T112, T122, T211, T212, T222);
                    computeMP(x1, y1, M1, M2, P11, P12, P22);
					gvar = IGW[l]*r*0.5;
                    
					U[k] += (-G11*f1 - G12*f2 +dEta*((T111*lnx+T112*lny)*u1+(T112*lnx+T122*lny)*u2))*gvar;
                    V[k] += (-G12*f1 - G22*f2 +dEta*((T211*lnx+T212*lny)*u1+(T212*lnx+T222*lny)*u2))*gvar;
                    P[k] += (-M1*f1 - M2*f2 +dEta*((P11*lnx+P12*lny)*u1+(P12*lnx+P22*lny)*u2))*gvar;
				}
			}
		}
# pragma omp barrier
	}
}



void MatBlC1::VisuSineFlow(int inP, double *Xpos, double *Ypos, double qAvel, double wavenumber, double dEta, int points, double *U, double *V, double *P){
	double dx, dy, xls, yls, xgs, ygs, r, lnx, lny, u1, u2, f1, f2;
	double x1, y1, gvar, G11, G12, G22, M1, M2;
	double T111, T122, T112, T211, T222, T212, P11, P12, P22;
	double dxpanel, dypanel, xppanel, yppanel, Vn, dpanel, nxpanel, nypanel, scoord;
	double *Xp, *Yp;
	int localElm = bem->allSize[inP];
    
	int xMatID = 0;
    
	
	Xp = bem->allXId[inP];
    Yp = bem->allYId[inP];
    
	for(int m=0; m<inP; m++){
		xMatID += 2*bem->allSize[m];
	}
	
	// Determine Panel Parameters
	dypanel = Yp[localElm]-Yp[0];
	dxpanel = Xp[localElm]-Xp[0];
	xppanel = 0.5*(Xp[localElm]+Xp[0]);
	yppanel = 0.5*(Yp[localElm]+Yp[0]);
	dpanel = sqrt(dxpanel*dxpanel+dypanel*dypanel);
    nxpanel = dypanel/dpanel;
    nypanel = -dxpanel/dpanel;
	
	//	Size of the Element, the Position of its center, normals
	for(int n=0; n<localElm; n++){
		dx = Xp[n+1]-Xp[n];
        dy = Yp[n+1]-Yp[n];
        xls = 0.5*(Xp[n+1]+Xp[n]);
        yls = 0.5*(Yp[n+1]+Yp[n]);
        r = (dx*dx+dy*dy);
        r = sqrt(r);
        lnx = dy/r;
        lny = -dx/r;
        
        scoord = -(xls-Xp[0])*nypanel + (yls-Yp[0])*nxpanel;
        Vn = qAvel*cos(2*M_PI*wavenumber*scoord/dpanel);
        
		u1 = Vn*nxpanel;
        u2 = Vn*nypanel;
        f1 = rhs[xMatID+2*n];
        f2 = rhs[xMatID+2*n+1];
        
#pragma omp parallel
        {
#pragma omp for private(xgs,ygs,x1,y1,G11,G12,G22,T111,T112,T122,T211,T212,T222,gvar,M1, M2, P11,P12,P22 )
            //	Loop over all panels, switch for singularity position (midpoint corner)
            for(int k=0; k<points; k++){
                xgs = xls-Xpos[k];
                ygs = yls-Ypos[k];
				// Loop over Gauss points
				for(int l=0; l<6; l++){
					x1 = 0.5*dx*IGP[l]+xgs;
					y1 = 0.5*dy*IGP[l]+ygs;
                    
					computeGT(1, x1, y1, G11, G12, G22, T111, T112, T122, T211, T212, T222);
                    computeMP(x1, y1, M1, M2, P11, P12, P22);
					gvar = IGW[l]*r*0.5;
                    
					U[k] += (-G11*f1 - G12*f2 +dEta*((T111*lnx+T112*lny)*u1+(T112*lnx+T122*lny)*u2))*gvar;
                    V[k] += (-G12*f1 - G22*f2 +dEta*((T211*lnx+T212*lny)*u1+(T212*lnx+T222*lny)*u2))*gvar;
                    P[k] += (-M1*f1 - M2*f2 +dEta*((P11*lnx+P12*lny)*u1+(P12*lnx+P22*lny)*u2))*gvar;
				}
			}
		}
# pragma omp barrier
	}
}

void MatBlC1::VisuCircTrac(int inP, double *Xpos, double *Ypos, int points, double *U,
    // Only for consant surface tension and delT = deltaT0 ou dT=0
    double *V, double *P){
	double dx, dy, xls, yls, xgs, ygs, r, lnx, lny;
	double x1, y1, G11, G12, G22, fl1, fl2, fr1, fr2;
	double T111, T122, T112, T211, T222, T212,ul,vl, ur, vr, f1, f2;
	double phil, phir, tx1, ty1, tx3, ty3, s1, s3, d1, d2, unorm, P11, P12, P22, M1, M2;
	double dLambda = Lambda -1.0;
    double *Xp, *Yp;
	int localElm = bem->allSize[inP];
	int nm1, np1, np2;
	int xMatID = 0;
    double LMgam, LPgam, RMgam,RPgam;
    double urr,vrr,ull,vll, kappa;
    double alpha = 1.0;
    if(kF < 1e-3){
		alpha=0;
	}
    
unorm= 0.0;
    Xp = bem->allXId[inP];
    Yp = bem->allYId[inP];
    
    
	
	for(int m=0; m<inP; m++){
		xMatID += 2*bem->allSize[m];
	}
	
	//	Size of the Element, the Position of its center, normals
	for(int n=0; n<localElm; n++){
		nm1 = n-1;
		np1 = n+1;
		np2 = n+2;
		if (n==localElm-1) {
			np1 = 0;
			np2 = 1;
		} else if(n==0){
			nm1 = localElm-1;
		}else if(n==localElm-2){
			np2=0;
		}
		
		tx1 = Xp[n]-Xp[nm1];
		ty1 = Yp[n]-Yp[nm1];
		s1 = sqrt(tx1*tx1+ty1*ty1);
		tx3 = Xp[np2]-Xp[np1];
		ty3 = Yp[np2]-Yp[np1];
		s3 = sqrt(tx3*tx3+ty3*ty3);
		dx = Xp[np1]-Xp[n];
		dy = Yp[np1]-Yp[n];
		r = sqrt(dx*dx+dy*dy);
		d1 = s1+r;
		d2 = s3+r;
		
        xls = 0.5*(Xp[n+1]+Xp[n]);
        yls = 0.5*(Yp[n+1]+Yp[n]);
        lnx = dy/r;
        lny = -dx/r;
        
        LMgam = LocalGam(inP, nm1, n)-LocalGam(inP, n, n)*(1.0-M_PI_4)*alpha;
        LPgam = LocalGam(inP, n, np1)-LocalGam(inP, n, n)*(1.0-M_PI_4)*alpha;
        RMgam = LocalGam(inP, n, np1)-LocalGam(inP, np1, np1)*(1.0-M_PI_4)*alpha;
        RPgam = LocalGam(inP, np1, np2)-LocalGam(inP, np1, np1)*(1.0-M_PI_4)*alpha;
        
		fl1 = 2.0*(LPgam*dx/r-LMgam*tx1/s1)/d1;
		fl2 = 2.0*(LPgam*dy/r-LMgam*ty1/s1)/d1;
		fr1 = 2.0*(RPgam*tx3/s3-RMgam*dx/r)/d2;
		fr2 = 2.0*(RPgam*ty3/s3-RMgam*dy/r)/d2;
        
        kappa = LocalGam(inP, n, n);
        (this->*applybc)(Xp[n], Yp[n], lnx, lny, unorm, kappa, f1, f2);
		fl1 +=  f1;
        fl2 +=  f2;

        kappa = LocalGam(inP, np1, np1);
        (this->*applybc)(Xp[n+1], Yp[n+1], lnx, lny, unorm, kappa, f1, f2);
		fr1 +=  f1;
        fr2 +=  f2;
        
        ul = rhs[xMatID+2*n];
        vl = rhs[xMatID+2*n+1];
        
        ull = rhs[xMatID+2*nm1];
        vll = rhs[xMatID+2*nm1+1];
        
        ur = rhs[xMatID+2*np1];
        vr = rhs[xMatID+2*np1+1];
        
        urr = rhs[xMatID+2*np2];
        vrr = rhs[xMatID+2*np2+1];
		
        fl1 += SpecialForces(inP,n)*lnx;
        fl2 += SpecialForces(inP,n)*lny;
        fr1 += SpecialForces(inP,np1)*lnx;
        fr2 += SpecialForces(inP,np1)*lny;
        
		//	Loop over all panels, switch for singularity position (midpoint corner)
#pragma omp parallel
        {
#pragma omp for private(xgs,ygs,x1,y1,G11,G12,G22,T111,T112,T122,T211,T212,T222,phil,phir,M1, M2, P11,P12,P22)
            for(int k=0; k<points; k++){
                xgs = xls-Xpos[k];
                ygs = yls-Ypos[k];
                
                // Loop over Gauss points
				for(int l=0; l<6; l++){
					x1 = 0.5*dx*IGP[l]+xgs;
					y1 = 0.5*dy*IGP[l]+ygs;
                    
					computeGT(1, x1, y1, G11, G12, G22, T111, T112, T122, T211, T212, T222);
                    computeMP(x1, y1, M1, M2, P11, P12, P22);
                    
					phil = (1.0-IGP[l])*0.25*IGW[l]*r;
					phir = (1.0+IGP[l])*0.25*IGW[l]*r;
                    
                    U[k] += -(fl1*G11+fl2*G12)*phil-(fr1*G11+fr2*G12)*phir
                    +dLambda*(phil*ul+phir*ur)*(T111*lnx+T112*lny)+dLambda*(phil*vl+phir*vr)*(T112*lnx+T122*lny);
                    V[k] += -(fl1*G12+fl2*G22)*phil-(fr1*G12+fr2*G22)*phir
                    +dLambda*(phil*ul+phir*ur)*(T211*lnx+T212*lny)+dLambda*(phil*vl+phir*vr)*(T212*lnx+T222*lny);
                    P[k] += -(fl1*M1+fl2*M2)*phil-(fr1*M1+fr2*M2)*phir
                    +dLambda*(phil*ul+phir*ur)*(P11*lnx+P12*lny)+dLambda*(phil*vl+phir*vr)*(P12*lnx+P22*lny);
                    
                    
                    phil = deltaT0*phil;
                    phir = deltaT0*phir;
                    
                    // Implicit Interface
                    U[k] += G11*phil/(s1*d1)*ull+G12*phil/(s1*d1)*vll;
					U[k] += -G11*(phil*(1.0/(s1*d1)+1.0/(r*d1))-phir/(r*d2))*ul-G12*(phil*(1.0/(s1*d1)+1.0/(r*d1))-phir/(r*d2))*vl;
					U[k] += -G11*(phir*(1.0/(r*d2)+1.0/(s3*d2))-phil/(r*d1))*ur-G12*(phir*(1.0/(r*d2)+1.0/(s3*d2))-phil/(r*d1))*vr;
					U[k] += G11*phir/(s3*d2)*urr+G12*phir/(s3*d2)*vrr;

					V[k] += G12*phil/(s1*d1)*ull+G22*phil/(s1*d1)*vll;
					V[k] += -G12*(phil*(1.0/(s1*d1)+1.0/(r*d1))-phir/(r*d2))*ul-G22*(phil*(1.0/(s1*d1)+1.0/(r*d1))-phir/(r*d2))*vl;
					V[k] += -G12*(phir*(1.0/(r*d2)+1.0/(s3*d2))-phil/(r*d1))*ur-G22*(phir*(1.0/(r*d2)+1.0/(s3*d2))-phil/(r*d1))*vr;
					V[k] += G12*phir/(s3*d2)*urr+G22*phir/(s3*d2)*vrr;
					
				}
			}
		}
# pragma omp barrier
	}
}

// void MatBlC1::VisuBordTrac(int inP, double *Xpos, double *Ypos, int points, double *U, double *V, double *P){

void MatBlC1::VisuInfiniBox(double *Xpos, double *Ypos, double uvel, double vvel, int points, double *U, double *V, double *P, int *Dom){
#pragma omp parallel
    {
#pragma omp for
        for(int k=0; k<points; k++){
            U[k] += 4*M_PI*uvel;
            V[k] += 4*M_PI*vvel;
            P[k] += -kF*kF*(uvel*Xpos[k]+vvel*Ypos[k])*2*M_PI;
		}
    }
}

void MatBlC1::VisuLinearFlow(double *Xpos, double *Ypos, double lin, double colin, int points, double *U, double *V, double *P, int *Dom){
#pragma omp parallel
    {
#pragma omp for
        for (int k=0; k<points; k++) {
            U[k] += 4*M_PI*(lin*Xpos[k] + colin*Ypos[k]);
            V[k] += 4*M_PI*(-lin*Ypos[k] + colin*Xpos[k]);
            P[k] += -0.5*kF*kF*(lin*(Xpos[k]*Xpos[k]-Ypos[k]*Ypos[k])+colin*Xpos[k]*Ypos[k])*2*M_PI;
        }
# pragma omp barrier
    }
}

void MatBlC1::VisuInfiniSource(double *Xpos, double *Ypos, double qdebit, double xpos, double ypos, double dEta, int points, double *U, double *V, double *P){
    double R;
#pragma omp parallel
    {
#pragma omp for	private(R)
        for(int k=0; k<points; k++){  
            R = Xpos[k]*Xpos[k]+Ypos[k]*Ypos[k];
            U[k] += 4*M_PI*qdebit*Xpos[k]/R;
            V[k] += 4*M_PI*qdebit*Ypos[k]/R;
            P[k] += -kF*kF*2*M_PI*qdebit*log(R);
		}
# pragma omp barrier
    }
}
