// ## Validating convergence on a Marangoni flow 
/**
    Test the convergence with the analytical solution of Boos and Thess
    Date: 09.12.2014
    Version 0.7
*/

// File Header
#include <iostream>
#include <stdio.h>
#include "../source/matblc1.h"    // provides the matrix object
#include "../source/bndblock.h"   // provides the boundary object
#ifndef _WIN32
#include <sys/time.h>
#endif
#ifdef _WIN32
#define _USE_MATH_DEFINES // for C++
#include <cmath>
#endif

#include <time.h>
#include <math.h>
int main (int argc, char * const argv[]) {
	
    std::cout << "Welcome to ULAMBATOR++ v.7\n   Unsteady LAminar Microfluidic Boundary element Algorithm To Obtain Results\n";
	char* savedir;
	
// Initial definitions
    
    // Where to store the output. Make sure that the folder exists.
    savedir = new char[120];
	sprintf(savedir,".");
    
    // One boundary is created
    BndBlock bem = BndBlock(1);
    
    // The channel aspect ratio L/H, the non-dimensional channel height is 1/LH.
    double LH = 8.0;
    
    // The viscosity ratio mu_inside/mu_outside
    double lambda = 0.2;
    
    // The capillary number, the problem runs with a velocity scale such that Ca=1
    // providing the capillary number here changes the inflow velocity
    double Ca = 0.0;
    
    // The number of elements
    int num_elements = 100;
    
    // Here the program catches the numbers that were passed when it was called
    // eg. ./ulambator_boosthess 5 0.5 100 sets the LH=5, lambda=0.5 and number of elements = 100
    if (argc>3) {
        LH = atof(argv[1]);
        lambda = atof(argv[2]);
        num_elements = atoi(argv[3]);  // atoi because its integer
    }else{
	std::cout<<"Not enough parameters, require L/H, lambda, N_elements"<<std::flush<<"\n"; 
    }
    
    // Some output
    std::cout<<"Boos and Thess Convergence study R/H "<<LH<<" lambda "<<lambda<<" N "<<num_elements<<"\n";
    
    // boundary conditions usually take two values, here an array is created that contains the information
    double bcs[] = {0.0, 0.0};

    // position data for the boundaries takes up to six values, here an array is created
    double pos[6];

    // Boundary 1 is initialized, 1 panel with 2 times num_elements elements of type 2, type 2 means a droplet boundary,
    // Important initialize first all fixed boundaries and then all droplets
    bem.Elements[0].Init(1, 2*num_elements,2);

    // position array for a droplet contain center coordinate x = 0, y=0, radius=1,
    // sinosoidal perturbation amplitude = 0 and perturbation wavenumber = 1.
    pos[0] = 0.0; pos[1]= 0.0; pos[2]= 1.0; pos[3] = 0; pos[4] = 1;

    // The first panel of boundary 1 is initialized, has num_elements elements, it belongs to geometry type 2 (circle),
    // boundary condition 22 (Boos and Thess), bcs has no influence and pos.
    bem.PanelInit(0,num_elements, 2, 22, bcs, pos);

    // This point 1/1 boundaries are defined all panels are defined.
    // More information of different Panel types and boundary conditions is found in the Manual under PanelInit
    // This finishes the initilization of the BndBlock and we can create the matrix

    // A matrix is created with the provided geometry bem, the aspect ratio LH, the viscosity ratio and capillary number.
    // Ca has no influence on the computation here, it enters through the boundary conditons,
    // it is passed only for logging purposes, you can chose what number you want to pass
    MatBlC1 mat(&bem, LH, lambda, Ca);

    // Build the matrix, provide a timestep for stabilization (set deltaT=0 for no stabilization)
    mat.Build(0.0);
    // Solve the matrix to obtain stresses and velocities
    mat.Solve();
   
    mat.BoosThessPrintAnalytic();
    

// End of program
    std::cout << "Finished and Exiting\n"; 
    return 0;
}
