// ## Validation with a parallel channel flow
// A parallel flow in a rectangular channel is solved and compared to the analytical solution.
/*
Ulambator configuration file
Date: 09.12.2014
Version 0.7
*/

// ###File Header
// Together with included classes we add Ulambators `matblc1.h` and `bndblock.h`, which manage the geometry and generate the matrix that solves the posed problem.
#include "../source/matblc1.h"
#include "../source/bndblock.h"
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#ifndef _WIN32
#include <sys/time.h>
#endif
#ifdef _WIN32
#define _USE_MATH_DEFINES // for C++
#include <cmath>
#endif

#include <time.h>
#include <math.h>

// The program begins with `int main` and prints a greeting to the screen.
int main (int argc, char * const argv[]) {
    std::cout <<"Welcome to ULAMBATOR++ v.7\n";
    std::cout <<"Unsteady LAminar Microfluidic Boundary element Algorithm To Obtain Results\n";
	char* savedir;
	
// ###Defining the geometry
// The object `bem` is a boundary block, which works as a container that will manage all boundaries like fixed walls and liquid interfaces. Here we initialize the container to contain only a single boundary.
    int number_of_boundaries = 1;
    BndBlock bem = BndBlock(number_of_boundaries);
    
// Two arrays are created. Position data of the boundaries in `pos` takes up to six values and boundary conditions `bcs` usually take two values.
    double bcs[] = {1.0, 0.0};
    double pos[6];

// The first boundary, boundary 0, is initialized (C++ starts counting from 0 unlike Matlab). The boundary will consist of 4 panels and the number of discrete elements `num_elements` per panel is by default 30 per unit length. The boundary is of type 0, type 0 means fixed boundary. The default aspect ratio W/H of the channel, here WH = 6.
    int num_panels = 4;
    int num_elements = 30;
    int type_boundary = 0;
    double WH=6;
   
    if (argc>2) {
        WH = atof(argv[1]);
        num_elements = atoi(argv[2]);  // atoi because its integer
    }else{
        std::cout<<"Not enough parameters, require L/H, N_elements per unit length"<<std::flush<<"\n";
    }
    
    bem.Elements[0].Init(num_panels, 12*num_elements,type_boundary);

// The first panel of boundary 0 is initialized with `num_elements` elements and it belongs to geometry type 0, a straight panel. The boundary condition type is 3, a  Poiseuille inflow boundary condition. The array `bcs` has no x and y velocity and the `pos` array contains the start point x,y and end point x,y coordinates.
    pos[0] = 0.0; pos[1]= 1.0; pos[2]= 0.0; pos[3] = 0.0;
    int geo_type = 0;
    int bc_type = 3;
    bem.PanelInit(0,num_elements, geo_type, bc_type, bcs, pos);

// Panel 1. The boundary condition type is 0, a no-slip boundary condition. The last point of the former panel is the first point of the new panel.
    pos[0] = 0.0; pos[1]= 0.0; pos[2]= 5.0; pos[3] = 0.0;
    bc_type = 0;
    bem.PanelInit(0,num_elements*5, geo_type, bc_type, bcs, pos);

// Panel 2, with constant normal pressure boundary condition.
    pos[0] = 5.0; pos[1]= 0.0; pos[2]= 5.0; pos[3] = 1.0;
    bc_type = 9;
    bcs[0] = 0.0;
    bem.PanelInit(0,num_elements, geo_type, bc_type, bcs, pos);
    
// Panel 3, again with no-slip boundary condition.
    pos[0] = 5.0; pos[1]= 1.0; pos[2]= 0.0; pos[3] = 1.0;
    bc_type = 0;
    bem.PanelInit(0,num_elements*5, geo_type, bc_type, bcs, pos);

// ###Configuring the problem to solve.
// At this point all boundaries and all panels are defined. This finishes the initilization of the `BndBlock` and we can create the matrix object.
//    A matrix is created with the provided geometry bem, the aspect ratio `WH`, the viscosity ratio = zero and an unused number. A variable that represents the channel aspect ratio L/H, the non-dimensional channel height is 1/`WH`.
    MatBlC1 mat(&bem, WH, 0.0, 1.0);

// The directory savedir is passed to the matrix object (and automatically to the `Bndblock`). Make sure that the folder exists.
    // savedir = new char[120];
    // sprintf(savedir,".");
    // std::cout<<savedir<<"\n"<<std::flush;
    // mat.EnableWrite(savedir);
    
// Write all boundaries into the savedir (or subfolder) (here: bnd0pos.dat)
    // bem.allWrite();
    
// Build the matrix, provide a timestep for stabilization (set deltaT=0 for no stabilization). Then solve the matrix to obtain stresses and velocities.
    mat.Build(0.0);
    mat.Solve();
    
// Error estimation based the pressure at entry and one side wall. It's incomplete but to get an idea.
    double kF = sqrt(12.0)*WH;
    double p_ref = kF*kF*kF/(kF-2.0*tanh(kF/2));
    double Nerr = 0;
    for (int i=0; i<num_elements; i++) {
        Nerr += pow(mat.rhs[2*i] - 5*p_ref,2);
        
        Nerr += pow(mat.rhs[10*i+2*num_elements+1]-p_ref*(5.0-0.5*(bem.Elements[0].XId[1][5*i]+bem.Elements[0].XId[1][5*i+1])), 2);
        Nerr += pow(mat.rhs[10*i+2*num_elements+3]-p_ref*(5.0-0.5*(bem.Elements[0].XId[1][5*i+1]+bem.Elements[0].XId[1][5*i+2])), 2);
        Nerr += pow(mat.rhs[10*i+2*num_elements+5]-p_ref*(5.0-0.5*(bem.Elements[0].XId[1][5*i+2]+bem.Elements[0].XId[1][5*i+3])), 2);
        Nerr += pow(mat.rhs[10*i+2*num_elements+7]-p_ref*(5.0-0.5*(bem.Elements[0].XId[1][5*i+3]+bem.Elements[0].XId[1][5*i+4])), 2);
        Nerr += pow(mat.rhs[10*i+2*num_elements+9]-p_ref*(5.0-0.5*(bem.Elements[0].XId[1][5*i+4]+bem.Elements[0].XId[1][5*i+5])), 2);
    }
    Nerr = sqrt(Nerr)/(6*num_elements*p_ref);
    
    std::cout<<"Root mean square error of the pressure distribution "<<Nerr<<"\n";
    
// Prepare a sensor file sensor0.txt and write runtime, velocities and pressure for two different positions: x=1, y=0.1 and x=4, y=0.1 to sensor0.txt.
    // mat.SensorEnable(0);
    // mat.SensorWrite(0, 1.0, 0.1);
    // mat.SensorWrite(0, 4.0, 0.1);
    
    double x0, x1, y0, u0, v0, p0, p1;
    int zone0 = 0;
    x0 = 1.0;
    x1 = 4.0;
    y0 = 0.1;
    mat.BuildVisual(1, &x0, &y0, &u0, &v0, &p0, &zone0);
    mat.BuildVisual(1, &x1, &y0, &u0, &v0, &p1, &zone0);

// A last log of the time. Make sure all files are closed to avoid data loss.
    // mat.LogRuntime(0);
    // mat.DisableWrite();

// With the equations given in Nagel et al.[1] one can derive the pressure drop as: $\Delta p = - L \frac{k^3}{k-2\tanh(k/2)}$ with $k=\sqrt{12}WH$ and $L=3$.
    std::cout<<"Expected pressure drop from reconstruction: ";
    double k = sqrt(12.0)*WH;
    double dp = -3*pow(k,3.0)/(k-2*tanh(k/2.0));
    double err = fabs((p1-p0-dp)/dp);
    std::cout<<dp<<" and the actual pressure drop: "<<(p1-p0)<<"\n";
    std::cout<<"The relative error is: "<<err<<"\n";
    
// End of program
    std::cout << "Finished and Exiting\n"; 
    return 0;
}
